#!/usr/bin/env python3
import numpy as np
import pylab as pl
import scipy
import re
pl.ion()
#from astropy.io import fits as pf
from scipy.sparse import coo_matrix, dia_matrix, dok_matrix, diags
from scipy.optimize import leastsq, fmin_ncg, check_grad, approx_fprime
from saunerie.lsqfit import Chi2
from saunerie.fitparameters import FitParameters
from saunerie.bspline import BSpline, CardinalBSplineC, integ
from croaks import NTuple
from numpy.polynomial.legendre import leggauss
from sksparse.cholmod import cholesky, cholesky_AAt
from saunerie import optim, plottools
#from scipy.integrate import quad
import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.DEBUG)


from astropy.modeling import models, fitting
from scipy.integrate import simps
from astroquery.simbad import Simbad
import matplotlib.gridspec as gridspec

import new_fitmodel as nf

# first model :
# $$ y_{sn,t_{mjd}} = f_{sn} \times S\left(\frac{t_{mjd} - t_{max}}{1+s_{sn}}; \theta_{splines} \right) $$

theta_range = [-35, 35]

customSimbad = Simbad()
customSimbad.remove_votable_fields('coordinates')
customSimbad.add_votable_fields('sptype')

def sqp_cholm(model, var_model, y, h, x0, g0, n_iter=30, sigma=None, ll=0.001, dchi2=0.01):
    """ Solves a non-linear programming problem with equality constraints using a Gauss-Newton approximation of the Hessian
    The problem must write as follows:                                                                                                                                                                                Find x that minimize: \sum (y_i - model(x)_i)^2
    subject to the constraint h(x) = 0

    Parameters:
    -----------
    model: callable with the signature (params, jac=False)
           must return if jac: the model evaluated at point params
                         else: the model and its jacobian under the form of a sparse matrix
           can raise ValueError in which case the model is assumed undefined at x
    y: array
       The fit data
    h: callable with the same signature as model
    x0: array
        starting point
    n_iter: maximal number of iterations

    Returns:
    --------
    The best fit after n_iter iterations
    """
    n = len(x0)
    m = len(g0)
    nn = n+m
    x = x0
    g = g0
    iter = 0
    alpha = 0.7

    while iter < n_iter:
        print(iter)
        val, jacobian = model(x, jac=True)
        V, dV = var_model(g, val, x, jacobian=jacobian)
        residuals = y - val
        
        if sigma is not None:
            #jacobian.data /= sigma[jacobian.row]
            #residuals /= sigma
            variance = np.diag(V) + sigma**2
        else :
            variance = np.diag(V)
        w = coo_matrix(np.diag(1/variance))

        obj_start = (np.diag(w.toarray()) * residuals**2).sum() + len(residuals) * np.log(np.pi) + np.log(1./np.diag(w.toarray())).sum()
        
        B = hessian(w, x, g, jacobian, dV)
        R = model.regulz(ll, g)
        B += R
        gradf = - grad(residuals, w, jacobian, dV, x, g)
        
        if h is not None:
            cons, H = h(x, Nsn = len(g0), jac=True, mask=None)
            M = scipy.sparse.bmat([[B, H.T], [H, None]])
            #d0 = scipy.sparse.linalg.spsolve(M, -np.r_[gradf, cons])
            try:
                factor = cholesky(M, mode='simplicial')
            except:
                return x, B, M
            d0 = factor(-np.r_[gradf, cons])
            mu0 = d0[nn:]

            d0 = d0[:nn]
            def phi(vect):
                p = vect[:n]
                g = vect[n:]
                val = model(p)
                V = var_model(g, val, p)
                variance = np.diag(V) + sigma**2
                w = coo_matrix(np.diag(1/variance))
                return (w * ( y - val)**2).sum() + len(y) * np.log(np.pi) + np.log(1./np.diag(w.toarray())).sum() #((w * ( y - val))**2).sum()
            def crit(t):
                return phi(np.r_[x,g] + t * d0)

        t, fval, ni, funcalls = scipy.optimize.brent(crit, brack=(0, 1), full_output=True)
        dof = len(y) - nn
        print(('stepsize: %g, objective: %g -> %g, decrement: %g, D.o.F.: %d, objective/dof: %g' % (t, obj_start, fval, obj_start - fval, dof, fval / dof)))
        decrement = obj_start - fval
        print('g =',g)
        #print('dg =',d0[n:])
        x = x + t * d0[:n]
        g = g + t * d0[n:]
        print('constraint 1 : %f' % ((H[0] * np.r_[x,g]).sum()))
        print('constraint 2 : %f' % ((H[1] * np.r_[x,g]).sum()))
        if len(h.val) == 3:
            print('constraint 3 : %f' % ((H[2] * np.r_[x,g]).sum()))
        iter = iter + 1
        #345(4)
        if decrement < dchi2:
            break

    return x, g, B, M




def hessian(w, p, g, J, dV):
    """
    Create Hessian matrix needed to minimized the chi2, in order to fit models of light curves and variances on data, using Newton-Raphson's method.
    Appearing on the left hand side of the equation.
    Input : - w = Array, is the matrix in which each diagonal element is the inverse variance of the measurement ( size : len(x) * len(x)).
            - p = Array, is the parameter of the LIGTH CURVE model.
            - g = Array, is the parameter of the VARIANCE model.
            - J = coo_matrix, is the Jacobian of the Light curev model.
            - dV = coo_matrix, is the Jacobian of the variance model.
            - reml = bool, if the reml technic is needed
    Output : - H = coo_matrix, sparse hessian matrix
    """
    H0 = 2 * J.T * w * J
    H = coo_matrix(H0)
    H00 = H.data
    i00 = H.row
    j00 = H.col

    def trace(i, j, w = w, dV = dV):
        """
        Calculate the trace term of the product of 2 derivative (wrt the parameters i and j)  appearing in the equation.
        Input : - i = int, parameters of which we want to take de derivative.
                - j = int, parameters of which we want to take de derivative.
                - w = Array, is the matrix in which each diagonal element is the inverse variance of the measurement ( size : len(data) * len(data)).
                - dV = coo_matrix, is the Jacobian of the variance model.
                - reml = bool, if the reml technic is needed
        Output : - float, result of the trace.
        """
        dV = dV.toarray()
        #dVi = coo_matrix(np.diag(dV[i]))
        #dVj = coo_matrix(np.diag(dV[j]))
        dVi = dV[i]
        dVj = dV[j]
        #return (w * dVi * w * dVj).toarray().trace()
        wdiag = np.diag(w.toarray())
        return (wdiag*dVi*wdiag*dVj).sum()

    npar = p.shape[0] + g.shape[0]
    Tr = np.array([trace(i,j) for i in range(npar) for j in range(i, npar)])
    ii = [i for i in range(npar) for j in range(i, npar)]
    jj = [j for i in range(npar) for j in range(i, npar)]

    H_l = coo_matrix((np.hstack((Tr,Tr)), (np.hstack((ii,jj)), np.hstack((jj, ii)))), shape = ( npar, npar))
    H_l = coo_matrix((np.hstack((H_l.data, -np.diag(H_l.toarray())/2.)), (np.hstack((H_l.row, range(H_l.toarray().shape[0]))), np.hstack((H_l.col, range(H_l.toarray().shape[0]))))), shape = (npar,npar))
    H = coo_matrix((H00, (i00, j00)), shape = ( npar, npar))
    return H + H_l

def grad(res, w, J, dV, p, g):
    """
    Create gradient vector needed to minimized the chi2, in order to fit models of light curves and variances on data, using Newton-Raphson's method.                                                             
    Appearing on the right hand side of the equation.
    Input : - res = Array, is the residual of the measurement substracted by the model.            
            - w = Array, is the matrix in which each diagonal element is the inverse variance of the measurement ( size : len(x) * len(x)).
            - J = coo_matrix, is the Jacobian of the Light curev model.
            - dV = coo_matrix, is the Jacobian of the variance model.
            - p = Array, is the parameter of the LIGTH CURVE model.
            - g = Array, is the parameter of the VARIANCE model.           
    Output : - (-1) * resul = Array, the needed gradient vector.
    """
    def grad_theta(i, J = J, res = res, dV = dV, w = w):
        """
        Calculate the term i of the gradient vector.
        Input : - i = int, represent the i-th parameter term of the gradient.
        Output : - Float, the ith term
        """
        grad = (w.toarray().dot(np.diag(dV.toarray()[i]))).trace() - np.linalg.multi_dot([res.T, w.toarray(), np.diag(dV.toarray()[i]), w.toarray(), res])
        if i < p.shape[0]:
            grad += - 2 * np.linalg.multi_dot([J.toarray().T[i], w.toarray(), res])
        return grad 
    resul = np.array([grad_theta(i) for i in range(p.shape[0] + g.shape[0])])
    return - resul




class SModel(object):

    def __init__(self, xdata, rng_spline = theta_range, grid_size = 10):
        self.nb_sn = len(np.unique(xdata['id']))
        self.spline_basis = CardinalBSplineC(grid_size, rng_spline)
        self.xdata = xdata
        self.n_theta = grid_size
        self.pars = self.para_init(xdata)

    def para_init(self, xdata, plot = False):
        fp = [("flux", self.nb_sn), ("tmax", self.nb_sn), ("s", self.nb_sn), ("theta", len(self.spline_basis))]
        fp = FitParameters(fp)

        def fitpoly(x, y, yunc , deg, plot = True):
            fit = fitting.LevMarLSQFitter() #Linear
            poly_init = models.Polynomial1D(degree = deg)
            fitted_fn = fit(poly_init, x, y)#, weights=yunc)
            if plot:
                pl.figure()
                pl.plot(x, y, 'k.', label='Data')
                pl.plot(x, fitted_fn(x), 'b.', label='Fitted Model')
                pl.xlabel('x')
                pl.ylabel('y')
                pl.legend()
            return fitted_fn
        if plot:
            pl.figure()
        N = len(np.unique(xdata['id']))
        data = xdata #generate_Lc(N, plot = False)                                                                                                                                                                 
        nb_pt = 1000   #
        X_re = np.array([])
        Y_re = np.array([])
    
        for i in np.unique(data['id']):
            d = data[data['id']==i]
            #if np.where(d == d.max())[0] == 0   :
             #   break
            
            x = d['Date']
            y = d['Flux']
            yerr = d['FluxErr']
            pfit = fitpoly(x, y, yerr, 5, plot = False)
            rng = np.linspace(x.min(),x.max(),nb_pt)
            p = pfit(rng)
            m = p.max()
            t = rng[np.where(p == m)[0]][0]

            l = rng[np.where(p > m/2.)[0]]
            s = (l.max() - l.min())/(2*np.sqrt(2*np.log(2)))

            fp['flux'].full[i] = m
            fp['tmax'].full[i] = t
            fp['s'].full[i] = s #np.sqrt(mod2 - mod0**2)                                                                                                                                                                       
            X_re = np.hstack((X_re, (x-t)/((1+s)*(1+d['ZP']))))
            Y_re = np.hstack((Y_re, y/m))
            if plot:
                pl.plot(X_re, Y_re, 'r.')
        if plot:
            X, Y, Yerr = plottools.binplot(X_re, Y_re, nbins = 50, marker='p', color='Darkblue')

        spline = self.spline_basis

        J = spline.eval(X_re) 
        H = J.T * J + 0.01 * diags(np.ones(J.shape[1]))
        factor = cholesky(H.tocsc())
        theta = factor(J.T * Y_re)
        fp['theta'].full[:] = theta

        I1 = (theta * integ(spline)).sum()        
        I2 = (theta * integ(spline, n=1)).sum()
        I3 = (theta * integ(spline, n=2)).sum()
        print('first constraint int =', I1)
        print('second constraint int =', I2)
        print('third constraint int =', I3)

        fp['theta'].full /= I1
        fp['flux'].full *= I1

        return fp #, I1, I2

    
    def regulz(self, ll, g):
        """
        """
        n = len(self.pars.free) + len(g)
        d = np.zeros(n)
        d[:] = 1.
        d[self.pars.indexof('theta')] = 1.
        #        r = ll * diags(d)
        return ll * diags(d)


    def __call__(self, p, jac=False, xdata=None):
        if xdata is None:
            xdata = self.xdata
        #xdata = self.tdata
        self.pars.free = p
        ide = xdata['id']
        sp1 = (1. + self.pars['s'].full[ide])
        phases = (xdata['Date'] - self.pars['tmax'].full[ide]) / (sp1 * (1 + xdata['ZP'])) 
                
        flx = self.pars['flux'].full[ide]
        J = self.spline_basis.eval(phases)
        v = flx * (J * self.pars['theta'].full)

        if jac:
            N = len(xdata)
            n = len(self.spline_basis)
            isn = np.arange(N)

            inter_t = xdata[xdata['id']==0].shape[0]
            dB = self.spline_basis.deriv(phases)

            # derivée flux:                                                                                                                                                                                        
            ii_f = isn
            jj_f = self.pars['flux'].indexof(ide)
            val_f = J * self.pars['theta'].full

            #derivee tmax:                                                                                                                                                                                         
            ii_t = isn
            jj_t = self.pars['tmax'].indexof(ide)
            val_t = self.pars['flux'].full[ide] * (-1/(sp1 * (1 + xdata['ZP']))) * (dB * self.pars['theta'].full)

            #derivee s:                                                                                                                                                                                            
            ii_s = isn
            jj_s = self.pars['s'].indexof(ide)
            val_s = self.pars['flux'].full[ide] * (-phases/sp1) * (dB * self.pars['theta'].full)

            #derivee theta:                                                                                                                                                                                        
            ii_theta = J.row
            jj_theta = self.pars['theta'].indexof(J.col)
            val_theta = self.pars['flux'].full[ide[J.row]] * J.data

            i = np.hstack((ii_f, ii_t, ii_s, ii_theta))
            j = np.hstack((jj_f, jj_t, jj_s, jj_theta))
            val_jac = np.hstack((val_f, val_t, val_s, val_theta))

            idx_fix = np.where(j != -1)
            Jac = coo_matrix((val_jac[idx_fix], (i[idx_fix],j[idx_fix])), shape = ( N, len(self.pars.free))) #3 * self.nb_sn + self.nb_theta))                                                                     
            return  v , Jac
        else :
            return v


class VModel(object):
    """ 
    Two variance model can be implemented.
    variance model:  - variance = v **2 * flux**2, here v is a parameter and flux is the measure of the SN light curve
                     - variance = v **2 where the variance does not depend of the SN flux

    There are two models, labelled by the input "lab", 'cst' stands for constant and refers to the second one
    and anything else stands for 'dependant' and refers to the first one.
    """
    def __init__(self, n_theta, data, lab = 'cst'):
        """
        Initialisation of the model.
        Input : n_theta = Int, is the number of spline use in the model of light curves
                data = NTuple, is the light curve data
                lab = str, is the type of variance model wanted
        Output : Varience model
        """
        #self.value = model_sn                                                                                                                                                                                     
        self.data = data
        self.nb_sn = len(np.unique(data['id']))
        self.nb_theta = n_theta
        self.pars = self.para_init()
        self.lab = lab

    def para_init(self):
        """                                                                                                                                                                                                               Initialisation of the parameters of the model.
        the parameter v represent the standard deviation : sigma = np.sqrt(variance), one per Supernova
        Input :
        Output : FitParameters, class of parameter
        """
        fp = [("v", self.nb_sn)]
        fp = FitParameters(fp)
        #fp['v'].fix(0, val = 0.1 )
        fp['v'] = 0.075 #p['tmax'].fix(0)
        return fp

    def __call__(self, g, value, p, J=None):
        """
        Return the evalutaion of the model given a set of parameter p.

        Input : - value = NTuple, is the value of the light curve model
                - p = Array, is the VARIANCE MODEL parameters
                - J = Array, is the matrix of the derivation of the LIGHT CURVE MODEL.

        Output : - if J is given : Array, a diagonal matrix of the values of the variance model evaluate with the given parameter.
                 - else : - Array the same matrix as describe above
                          - dV = coo_matrix , sparse matrix of the derivative of the VARIANCE MODEL.
        """
        ide = self.data['id']

        self.pars.free = g

        if J == None:
            if self.lab == 'cst':
                return np.diag(self.pars['v'].full[ide]**2) #np.diag(p[ide])
            else :
                return np.diag((self.pars['v'].full[ide]*(value))**2) #np.diag(p[ide]*(value)**2)
        else :
            N = len(p) + len(g)
            n = len(value)
            isn = np.arange(n)

            # derive parametre de J:
            ii_J = J.col
            jj_J = J.row
            
            if self.lab == 'cst':
                val_J = np.zeros((len(J.data))) #2 * value[ide[J.row]] * J.data * p[ide[J.row]]
            else :
                val_J = 2 * value[ide[J.row]] * J.data * self.pars['v'].full[ide[J.row]]**2
            # derive V_{n}:

            ii_v = len(p) + ide #J.shape[1] + ide #len(p) + ide ######################################################################################################<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
            jj_v = isn
            if self.lab == 'cst':
                val_v = np.ones(len(value)) #value**2
            else :
                val_v = 2 * self.pars['v'].full[ide] * value**2

            i = np.hstack((ii_J, ii_v))
            j = np.hstack((jj_J, jj_v))
            val_jac = np.hstack((val_J, val_v))

            dV = coo_matrix((val_jac, (i,j)), shape = (N, n))
            if self.lab == 'cst':
                return np.diag(self.pars['v'].full[ide]**2), dV #*(value)**2), dV
            else :
                return np.diag((self.pars['v'].full[ide]*(value))**2), dV




        
        
class IntegralConstraints():
    """
    """
    def __init__(self, pars, basis, vals=[1., 0., 10.], limits=None):
        """
        """
        self.basis = basis
        self.val = np.array(vals)
        #JJ = np.array([])
        #for i in range(len(vals)):
        self.JJ = np.vstack([integ(self.basis, i) for i in range(len(vals))]).T
                                 #integ(self.basis, 1),
                                 #integ(self.basis, 2)]).T
        #self.JJ = JJ.T
        self.reset_pars(pars)
        #        self.pars = pars                                                                                                                                                                                 
        #        s = pars._struct.slices['theta']
        
        if limits is not None:
            self._integral_limits(limits[0], limits[1], self.JJ)

    def reset_pars(self, pars):
        """                                                                                             
        """
        self.pars = pars
        n = len(self.pars.free)
        self.J = np.zeros((n,len(self.val)))
        i = self.pars.indexof('theta')
        idx = i>=0 #
        self.J[i[idx],:] = self.JJ[idx]
        
    def _integral_limits(self, xmin, xmax, J):
        """
        should be deprecated
        """
        # limit the constraints to a range in x    
        g = self.basis.grid
        gg = 0.5 * (g[1:] + g[:-1])
        ok = np.zeros(len(self.basis)).astype(bool)
        ok[self.basis.order-1:] = (gg>=xmin) & (gg<=xmax)
        J[~ok] = 0.
        self.ok = ok

    #    def init_pars(self):
    #        r = FitParameters([('LL', 3)])
    #        return r                                                                                                                                                                                              

    def __call__(self, p, Nsn, jac=False, mask=None):
        """
        """
        if mask is not None:
            J = self.J.T * mask
        else:
            J = self.J.T
            JJ = np.zeros((J.shape[0], J.shape[1] + Nsn ))
            for i in range(J.shape[0]):
                JJ[i][:J.shape[1]] = J[i]
                
        val = J.dot(p) - self.val

        if jac == False:
            return val

        return val, JJ




def generate_Lc(N, dx = 0.01, rng = [-70, 180], variance_sn = 0.05 ,plot=False, lab = 'dep'):
    """                                                                                                                                                                                                            
    """
    x,y,yerr = [], [], []
    ide = []
    for i in range(N):
        tmax = np.random.uniform(rng[0], rng[1])
        sig = np.random.uniform(1., 10.)
        norm = np.random.uniform(50., 1000.)

        npts = np.random.randint(10,20)
        
        xx = np.random.uniform(-3*sig, 3*sig, size=npts)
        #xx = np.linspace(-20, 20, npts)
        yy = norm * np.exp(-0.5 * (xx/sig)**2)
        xx += tmax
        yy_err = 10.
        if lab == 'cst' :
            yy += np.random.normal(scale=np.sqrt((variance_sn)**2) , size=len(yy))#np.random.normal(scale=yy_err, size=len(yy))
        else :
            yy += np.random.normal(scale=np.sqrt((variance_sn * yy)**2) , size=len(yy))
        x.append(xx)
        y.append(yy)
        ide.append(np.full(len(xx), i))
        yerr.append(np.full(len(xx), yy_err))

    x = np.hstack(x)

    # data set                                                                                                                                                                                                     
    data = NTuple(len(x),dtype=[('Date', float), ('Flux', float), ('FluxErr', float), ('ZP', float) , ('id',int)])
    data['Date'] = np.hstack(x)
    data['Flux'] = np.hstack(y)
    data['FluxErr'] = np.hstack(yerr)
    data['ZP'] = 0.01 
    data['id'] = np.hstack(ide)

    if plot :
        for i in np.unique(data['id']):
            idx = data['id'] == i
            tmax = data['Flux'][idx].max()
            phase = data['Date'][idx]
            pl.errorbar(phase, data['Flux'][idx], yerr=data['FluxErr'][idx], fmt='.')

    return data




def fit_sqp(data = None, N = 20, C = [1., 0., 1.], n_theta = 30, max_iter= 200, nomin = False, survey = 'csp', band = b'SWOPE::B', SNe = None, lab = 'dep', variance_sn = 0.05):
    if data is None :
        print(data)
        data = generate_Lc(N, plot = False, variance_sn = variance_sn, lab = lab)
        
    else :
        try:
            data = data[data['Filter'] == band]
        except:
            data = data
        #init_data, redu_data, SNe = data_cons(data = data, band = band, survey = survey, before = 7, after = 7 )
        
    #data = data[un]
    sig = data['FluxErr']
    
    #w = 1. / data['FluxErr']
    model = SModel(data, grid_size = n_theta)
    var_model = VModel(model.n_theta, data, lab = lab)
    
    p = model.pars.free
    g = var_model.pars.free
    
    #Initialilsation
    init = model(p, jac=0)
    fig = pl.figure()
    ax1 = pl.subplot(211)
    ax2 = pl.subplot(212, sharex = ax1)

    for i in np.unique(data['id']):
        d = data[data['id'] == i]
        date = np.linspace(d['Date'].min(), d['Date'].max())
        u = np.full(len(date), i).astype(int)
        z = np.full(len(date), d['ZP'][0]).astype(float)
        d = np.rec.fromarrays((date, z, u), names=['Date', 'ZP','id'])
        v = model(p, jac=0, xdata=d)
        ax1.plot(d['Date'], v, color=pl.cm.jet(i), ls='-')
       
        
    ax1.errorbar(data['Date'], data['Flux'], yerr=data['FluxErr'],color ='k', marker='.', ls='')
    ax1.set_title(survey + '  ' + band.decode('UTF-8') + '   INIT')
    ax1.set_ylabel('Flux') 

    ax2.errorbar(data['Date'], data['Flux']-init, yerr=data['FluxErr'], color ='k', marker='.', ls = '')
    ax2.plot(np.linspace(data['Date'].min(), data['Date'].max(), 1000), np.zeros(1000), ls = '--', color = 'red', linewidth = 0.5 ) 
    vmin = (data['Flux']-init).min() - 0.1
    vmax = (data['Flux']-init).max() + 0.1
    ax2.set_ylim(vmin, vmax)
    ax2.set_ylabel('Residuals')
    ax2.set_xlabel('Days')
    cons = IntegralConstraints(model.pars, model.spline_basis, vals=C, limits=(-32., 32.))

    p, g, B, M = sqp_cholm(model, var_model, data['Flux'], cons, p, g, n_iter=max_iter, sigma = sig)
   
    fig = pl.figure()
    ax1 = pl.subplot(211)
    ax2 = pl.subplot(212, sharex = ax1)

    phase = data['Date']
    #p = model.pars.free
    chi2 = []
    #snw = []
    #sn = sn_csp()
    pl.figure()
    ax1.errorbar(phase, data['Flux'], yerr = data['FluxErr'], marker='.', ls='', color='k')
    for i in np.unique(data['id']):
        d = data[data['id'] == i]
        date = np.linspace(d['Date'].min(), d['Date'].max())
        u = np.full(len(date), i).astype(int)
        z = np.full(len(date), d['ZP'][0]).astype(float)
        d_sn = np.rec.fromarrays((date, z, u), names=['Date','ZP','id'])
        v = model(p, jac=0, xdata=d_sn)
        ax1.plot(d_sn['Date'], v, color=pl.cm.jet(i), ls='-')
        ax1.plot(d_sn['Date'], v + 1*(np.sqrt((g[i] * v)**2)), color=pl.cm.jet(i), ls='--', linewidth = 0.5)
        ax1.plot(d_sn['Date'], v - 1*(np.sqrt((g[i] * v)**2)), color=pl.cm.jet(i), ls='--', linewidth = 0.5)

        col = np.random.randint(0,1000)/1000
        
        pl.plot((d['Date'] - p[len(np.unique(data['id']))+i])/ (1 + p[2*len(np.unique(data['id'])) + i]), d['Flux']/np.abs(p[i]), marker = '.', color = pl.cm.jet(col), ls= '' )

        val = model(p, jac=0, xdata=d)
        chi2.append(((d['Flux'] - val)**2/d['FluxErr']**2).sum()/len(d)) 
        #if chi2[len(chi2)-1] >= 10:
         #   ax1.plot(d_sn['Date'], v, color='r', ls='-')
          #  ax1.text(d_sn['Date'][int(len(d_sn['Date'])/2)], v[v == v.max()], SNe[i].decode('UTF-8'))

    pl.title(survey + '  ' + band.decode('UTF-8') + '  Light Curves corrected with fit parameters')
    pl.xlabel(r'$\frac{Days - DayMax_{SN}}{(1+s_{SN})(1+z)}$')
    pl.ylabel(r'$\frac{Flux_{SN}}{|Fitted Flux_{SN}|}$')
    ax1.set_title(survey + '  ' + band.decode('UTF-8') + '  FIT')
    ax1.set_ylabel('Flux')
    v = model(p, jac=0)
    
    ax2.errorbar(phase, data['Flux']-v, yerr = data['FluxErr'], color ='k', marker='.', ls = '')
    ax2.plot(np.linspace(data['Date'].min(), data['Date'].max(), 1000), np.zeros(1000), ls = '--', color = 'red', linewidth = 0.5 )
    ax2.set_ylim(vmin, vmax)
    ax2.set_ylabel('Residuals')
    ax2.set_xlabel('Days')

    """
    pl.figure()
    pl.title(survey + '  ' + band.decode('UTF-8') + '  Chi2 = f(sn)')
    pl.ylabel(r'$\chi^{2}}$/nb_pts')
    pl.xlabel('SN')
    pl.scatter(range(len(chi2)), chi2, marker = '.')
    #compt = 0
    for i in range(len(chi2)):
        if chi2[i] > 10.:
            cst = np.random.randint(0, 5)
            sn0 = SNe[i].decode('UTF-8')
            sn0 = sn0[:2] + ' ' + sn0[2:]
            result_table = customSimbad.query_object(sn0)
            pl.text(i-1, chi2[i] + 3, sn0)
            pl.text(i-1, chi2[i] ,result_table['SP_TYPE'][0].decode('UTF-8'))
            #compt += 1

    SNe_type = []
    for i in range(len(chi2)):
        sn0 = SNe[i].decode('UTF-8')

    
    idx = np.array(chi2).argsort()
    pl.figure()
    pl.xlabel(r'$\chi^{2}}$/nb_pts')
    pl.ylabel('SN')
    pl.title(survey + '  ' + band.decode('UTF-8') + '  SN = f(Chi2)')
    pl.plot(np.array(chi2)[idx], SNe[idx], 'k.')
    """

    #if B is not None:
        #fact = cholesky(B, mode='simplicial')
        #U = fact.inv().todense()

        #    U = np.linalg.inv(HH.todense())                                                                                                                                                                       
        #D = np.matrix(np.diag(U))
        #C = U / np.sqrt(D.T * D)
        #pl.matshow(C)
        #pl.colorbar()

    return p, g, model, var_model #s, data, model, cons, B, M#, snw

def control(real = 10, N = 10, sigma_err = 1., variance_sn = 0.05, lab = 'cst', dbin = 0.02, plot = False):
    
    """                                   
    Control plot : histogram of the variance parameters for found for all SNe.
    Input : - real = int, number of realisation of the fit.
            - N = int, number of SNe fitting in each "real" realisation.
            - sigma_err = float, measure error we want on the SNe light curves
     




       - variance_sn = float, variance we want to be modeled by the variance model.
            - lab = str 'cst' or 'dep', type of variance model we want to use.
            - nbins = int, number of bins for the histogram.
            - plot = bool, True if we want to see each Sne light curve plot with the foncion plot_cl.       Output : histogram
    """
    g = []
    for i in range(real):
        #data = generate_Lc(N = N, variance_sn = variance_sn)#, lab = lab)
        #data = generate_Lc(N, plot = False, variance_sn = variance_sn)
        p0, g0, model, var_model = fit_sqp(data = None)#, max_iter = 100, lab = lab)#, plot = plot)
        g.append(g0)

    g = np.array(g)
    pl.figure()

    xx, yy = np.histogram(g.ravel())
    rg = (yy.min(), yy.max())
    nbins = int((rg[1] - rg[0])/ dbin)

    xh, yh, hh = pl.hist(g.ravel(), bins=nbins)#, range = (,0.01))                                       

    med = np.median(g)
    moy = g.mean()
    pl.plot(variance_sn * np.ones(10), np.linspace(0,xh.max(),10), 'r--')
    pl.plot(moy * np.ones(10), np.linspace(0,xh.max(),10), 'k--')
    pl.plot(med * np.ones(10), np.linspace(0,xh.max(),10), 'b--')
    pl.title('%i real  de %i SN : v = %f ( => v_simu moy : %f med : %f)' % (real, N, variance_sn, moy, med))


def comparaison(N = 20, variance_sn = 0.05, lab = 'dep'):
    data = generate_Lc(N, plot = False, variance_sn = variance_sn, lab = lab)

    p, g, model, var_model = fit_sqp(data = data, lab = lab)
    
    data, model1, cons, B, M = nf.fit_sqp(data = data)

    phase = data['Date']
    p1 = model1.pars.free
    
    fig = pl.figure(figsize=(10, 10))
    outer = gridspec.GridSpec(2, 1, wspace=0.2, hspace=0.2)
    inner = gridspec.GridSpecFromSubplotSpec(2, 1, subplot_spec=outer[0], wspace=0.1, hspace=0.1)

    ax1 = pl.Subplot(fig, inner[0])
    ax2 = pl.Subplot(fig, inner[1], sharex = ax1)
    ax1.errorbar(phase, data['Flux'], yerr = data['FluxErr'], marker='.', ls='', color='k')
    for i in np.unique(data['id']):
        d = data[data['id'] == i]
        #idx_bads = bads[data['id'] == i]
        date = np.linspace(d['Date'].min(), d['Date'].max())
        u = np.full(len(date), i).astype(int)
        z = np.full(len(date), d['ZP'][0]).astype(float)
        d_sn = np.rec.fromarrays((date, z, u), names=['Date','ZP','id'])
        v = model(p, jac=0, xdata=d_sn)
        ax1.plot(d_sn['Date'], v, color=pl.cm.jet(i**2), ls='-')
        if lab == 'cst' :
            print('il entre ')
            ax1.plot(d_sn['Date'], v + 1*(np.sqrt((g[i])**2)), color=pl.cm.jet(i**2), ls='--', linewidth = 0.5)
            ax1.plot(d_sn['Date'], v - 1*(np.sqrt((g[i])**2)), color=pl.cm.jet(i**2), ls='--', linewidth = 0.5)
        else :
            ax1.plot(d_sn['Date'], v + 1*(np.sqrt((g[i] * v)**2)), color=pl.cm.jet(i**2), ls='--', linewidth = 0.5)
            ax1.plot(d_sn['Date'], v - 1*(np.sqrt((g[i] * v)**2)), color=pl.cm.jet(i**2), ls='--', linewidth = 0.5)
    ax1.set_ylabel('Flux')
    v = model(p, jac=0)
    ax2.errorbar(phase, data['Flux']-v, yerr = data['FluxErr'], color ='k', marker='.', ls = '')
    ax2.plot(np.linspace(data['Date'].min(), data['Date'].max(), 1000), np.zeros(1000), ls = '--', color = 'red', linewidth = 0.5 )
    #ax2.set_ylim(vmin, vmax)
    ax2.set_ylabel('Residuals')
    ax2.set_xlabel('Days')
    fig.add_subplot(ax1)
    fig.add_subplot(ax2)

    inner = gridspec.GridSpecFromSubplotSpec(2, 1, subplot_spec=outer[1], wspace=0.1, hspace=0.1)
    ax3 = pl.Subplot(fig, inner[0], sharex = ax1, sharey = ax1)
    ax4 = pl.Subplot(fig, inner[1], sharex = ax3, sharey = ax2)
    ax3.errorbar(phase, data['Flux'], yerr = data['FluxErr'], marker='.', ls='', color='k')
    for i in np.unique(data['id']):
        d = data[data['id'] == i]
        #idx_bads = bads[data['id'] == i]
        date = np.linspace(d['Date'].min(), d['Date'].max())
        u = np.full(len(date), i).astype(int)
        z = np.full(len(date), d['ZP'][0]).astype(float)
        d_sn = np.rec.fromarrays((date, z, u), names=['Date','ZP','id'])
        v = model1(p1, jac=0, xdata=d_sn)
        ax3.plot(d_sn['Date'], v, color=pl.cm.jet(i**2), ls='-')
    ax3.set_ylabel('Flux')
    v = model1(p1, jac=0)

    ax4.errorbar(phase, data['Flux']-v, yerr = data['FluxErr'], color ='k', marker='.', ls = '')
    ax4.plot(np.linspace(data['Date'].min(), data['Date'].max(), 1000), np.zeros(1000), ls = '--', color = 'red', linewidth = 0.5 )
    #ax4.set_ylim(vmin, vmax)
    ax4.set_ylabel('Residuals')
    ax4.set_xlabel('Days')
    fig.add_subplot(ax3)
    fig.add_subplot(ax4)



    
    
    
def data_cons(data = None, band = b'SWOPE::B', survey = 'csp', mjd_plus = 30, mjd_moins = 15):#,  n_pt = 7):
    if data is None:
        data = NTuple.fromfile('work/snls5/' + survey + '_lc.npy')
        
    data = data[data['Filter'] == band]
    mjd_max = np.zeros_like(data['Date'])
    
    for i in np.unique(data['id']):
        
        d = np.where(data['id'] == i)[0]
        #date = data[d]['Date']
        if len(d) < 6:
            #print('non', data[d]['id'])
            mjd_max[d] = 0
        else :
            #print('oui', data[d]['id'])
            dmax = data['Date'][d][np.where(data[d]['Flux'] == data[d]['Flux'].max() )[0]]
            mjd_max[d] = dmax.min()
            #print(data['Date'][d], dmax, mjd_max)
            #544(233)
        
    date_red = data['Date'] - mjd_max
    #234(23)
    #print(date_red)
    #for i_date in date-red:
        #if i_date > 0. :
            
    idx = (date_red < -mjd_moins) ^ (date_red > mjd_plus) 
    #3432(234)
    """if len(d) < 6:
            continue
        else:
            id_fmax = np.where(d['Flux'] == d['Flux'].max())[0][0]
            idx_sn = np.where(data['id']==d['id'][0])[0]
            
            if id_fmax != 0:
                if id_fmax >= n_pt:
                    for ii in idx_sn[id_fmax - n_pt : id_fmax + n_pt + 1]:
                        idx.append(ii)
                else:
                    for ii in idx_sn[: id_fmax + n_pt + 1]:
        idx.append(ii)
    """    
    data = data[~idx]

    len_data = np.zeros_like(data['Date'], dtype = bool)
    for i in np.unique(data['id']):
        d = np.where(data['id'] == i)[0]
        if len(d) < 6:
            len_data[d] = False
        else :
            len_data[d] = True

    data = data[len_data]

    
    #reduced data
    idin_sn = [] 
    for i in range(len(data)):
        if data['id'][i] in data['id']:
             idin_sn.append(i)
    un = np.unique(data['id'])                     
    for i in range(len(un)):
        idex0 = data['id'] == un[i]
        data['id'][idex0] = i    
    data['Flux'] *= 10**6 
    data['FluxErr'] *= 10**6    
    sn_sum = np.load('work/snls5/' + survey + '_lc_summary.npy')
    return data , sn_sum['SN'][un]#init_data, data[idin_sn], sn_sum['SN'][un]


def sn_csp():
    fi = open('work/snls5/studies/samples_selection/CSP/LC/DR3/tab1.dat')
    ligne = fi.readlines()

    keys = [('SN', '|S20'), ('ZHelio', float)]
    sn = np.zeros(len(ligne[1:-7]), dtype = keys)
    
    p = re.compile(r'\t+')
    dicti = dict()                                                       
    c =  0
    for i in ligne[1:-7]:
        v = p.split(i)
        sn[c] = v[0], v[5]
        c +=1
    return sn


def find_sn(survey = 'csp'):
    xdata = np.load('work/snls5/' + survey + '_lc.npy')
    filter = np.unique(xdata['Filter'])
    wi = np.zeros(len(filter), dtype = [('band', '|S20'), ('SNe', list)])
    c = 0                       
    for i in filter:
        print(i)
        data, sn,  un = data_cons(band = i)
        s, data, model, cons, B, M = fit_sqp(data = data, C = [1., 0., 10.], max_iter = 250, n_theta = 30, band=i)
        wi[c] = i, snw
        c += 1
    return wi



import numpy as np
import pylab as pl
from nacl.models.regularizations import NaClSplineRegularization
from nacl import minimize
import helpers

tds, model = helpers.generate_dataset(2, seed=42)


def test_regularization_minimization(to_regularize=['M0']):
    """
    """
    model.pars.fix()
    for block_name in to_regularize:
        model.pars[block_name].release()

    reg =NaClSplineRegularization(model, to_regularize=['M0'])
    minz = minimize.Minimizer(reg)
    minz(model.pars.free)
    return reg, minz






#!/usr/bin/env python3

import numpy as np
import matplotlib.pyplot as plt

from lemaitre import bandpasses
from nacl.models import helpers
from nacl.models import salt2
from nacl.models.salt2 import lightcurves
from nacl.models.salt2 import spectra
from nacl import TrainingDataset


def load(filename='snf_projected.hd5', keep=['SNF20051003-004', 'PTF10qyz', 'PTF11kly', 'SNF20080803-000']):
    """
    """
    fl = bandpasses.get_filterlib()

    if 'hd5' in filename:
        tds = TrainingDataset.read_hdf(filename)
    else:
        tds = TrainingDataset.read_parquet(filename, filterlib=fl)

    if keep is not None:
        sne = tds.sn_data.sn_set.tolist()
        for k in keep:
            sne.remove(k)
        tds.kill_sne(sne)
        tds.compress()

    return tds


def check_color_law_grad(color_law, wl, pars):
    """
    """
    pp = pars['CL'].full

    v, jacobian = color_law(wl, pp, jac=True)
    dx = 1.E-7
    ppp = pp.copy()
    df = []
    for i in range(len(pp)):
        ppp[i] += dx
        vp, _ = color_law(wl, ppp, jac=False)
        df.append((vp-v)/dx)
        ppp[i] -= dx
    return jacobian, np.vstack(df).T


if __name__ == '__main__':
    tds = load()
    model = salt2.SALT2Like(tds, wl_grid=np.linspace(2000., 11000., 202))
    pars = model.init_pars()
    wl = np.linspace(2200., 9200., 250)
    J, Jn = check_color_law_grad(model.color_law, wl, pars)

    fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(16,5))
    axes[0].imshow(J, aspect='auto')
    axes[0].set_title('color law derivatives - analytical')
    axes[1].imshow(Jn, aspect='auto')
    axes[1].set_title('color law derivatives - numerical')
    c = axes[2].imshow(J-Jn, aspect='auto')
    axes[2].set_title('analytical - numerical')
    plt.colorbar(c)

    fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(8,8))
    J = J.flatten()
    Jn = Jn.flatten()
    axes[0,0].plot(J, Jn, 'k.')
    axes[0,1].plot((J-Jn)/J, 'k.')
    axes[1,0].plot(J, J-Jn, 'k.')



import nacl.simulations.ideal as ideal
from nacl.models.salt import SALT2Like
from nacl.simulations.flux_simulator import Simulator
from nacl.models.variancemodels import SimpleVarianceModel
from nacl.minimizers import Minimizer
from nacl.fit import FitModel2D
import numpy as np 

N_SN = 500
MU_REG = 10

####################################### SIMULATION #############################
# Create the cadence 
DS = ideal.Generator(n_sn=N_SN, lsst_simu=True)

# Simulate the flux and the measurements error
INIT_FROM_SALT2_FILE = "../data/salt2.npz"

# Error snake initiate at 5%
GAMMA_INIT = 0.05  
FS = Simulator(DS,                                        # DataSet
               SALT2Like,                                 # Model function used to simulated
               variance_model_func=SimpleVarianceModel,   # Error Snake used in sim
               gamma_init=GAMMA_INIT,                     # Value of the error snake
               init_from_salt2_file=INIT_FROM_SALT2_FILE) # Model values for sim

####################################### FIT ####################################
## Instantiate Fit
# Create Constraints and Regularization
# as function of free parameters.
fit = FitModel2D(FS.trainingDataset,                            # TrainingDataset
                 model_func=SALT2Like,                          # Model used
                 normalization_band_name='SWOPE::B',            # Model normalization
                 init_from_salt2_file=INIT_FROM_SALT2_FILE,     # Files to initiate the model
                 pars_fix=[],                                   # Parameters that should be fixed
                 mu_reg=MU_REG,                                 # Regularization hyperparameter
                 minimizer=Minimizer)                           # Minimizer function

# degrade SN parameters (multiply by N(1, 0.05))
# degrade error snake  (multiply by N(1, 1))
fit.parameter_degradation(model_coef_init=0.05,
                          var_coef_init=0.1)

################################################### 1st fit model without Error Snake 
fit(pars_fix=['eta_calib', 'kappa_color', 'gamma'])

################################################### 2nd Fit
# Release Error Snake parameter
fit.variance_model.pars.release()
fit(pars_fix=fit.all_pars, g=fit.variance_model.pars.free)

################################################### 3rd Fit
# Release model parameters
fit.model.pars.release()
fit(pars_fix=['eta_calib', 'kappa_color'], g=fit.variance_model.pars.free)



"""
"""
import logging

import os.path as op
import ctypes
import numpy as np
import scipy.sparse

class Triplet(ctypes.Structure):
    _fields_ = [("row", ctypes.c_int),
                ("col", ctypes.c_int),
                ("value", ctypes.c_double)]


lib = np.ctypeslib.load_library("_libnacl", op.dirname(op.abspath(__file__)))
lib.kron_product_by_line.restype = ctypes.c_int
lib.kron_product_by_line.argtypes = [
    ctypes.c_int, ctypes.c_int, ctypes.c_int,
    np.ctypeslib.ndpointer(dtype=np.int32, ndim=1, flags='aligned, contiguous'),
    np.ctypeslib.ndpointer(dtype=np.int32, ndim=1, flags="aligned, contiguous"),
    np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags="aligned, contiguous"),
    np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags="aligned, contiguous"),

    np.ctypeslib.ndpointer(dtype=np.int32, ndim=1, flags='aligned, contiguous, writeable'),
    np.ctypeslib.ndpointer(dtype=np.int32, ndim=1, flags='aligned, contiguous, writeable'),
    np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags='aligned, contiguous, writeable'),
    # ctypes.POINTER(Triplet)
]


lib.kron_product_by_line_2.restype = ctypes.c_int
lib.kron_product_by_line_2.argtypes = [
    ctypes.c_int, ctypes.c_int, ctypes.c_int,
    np.ctypeslib.ndpointer(dtype=np.int32, ndim=1, flags='aligned, contiguous'),
    np.ctypeslib.ndpointer(dtype=np.int32, ndim=1, flags="aligned, contiguous"),
    np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags="aligned, contiguous"),
    np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags="aligned, contiguous"),

    np.ctypeslib.ndpointer(dtype=np.int32, ndim=1, flags='aligned, contiguous, writeable'),
    np.ctypeslib.ndpointer(dtype=np.int32, ndim=1, flags='aligned, contiguous, writeable'),
    np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags='aligned, contiguous, writeable'),
]


lib.append.restype = ctypes.c_int
lib.append.argtypes = [
    ctypes.c_int, ctypes.c_int,
    np.ctypeslib.ndpointer(dtype=np.int64, ndim=1, flags='aligned, contiguous'),
    np.ctypeslib.ndpointer(dtype=np.int64, ndim=1, flags="aligned, contiguous"),
    np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags="aligned, contiguous"),

    np.ctypeslib.ndpointer(dtype=np.int64, ndim=1, flags='aligned, contiguous, writeable'),
    np.ctypeslib.ndpointer(dtype=np.int64, ndim=1, flags='aligned, contiguous, writeable'),
    np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags='aligned, contiguous, writeable'),

    ctypes.c_int
]



def kron_product_by_line(A, B):
    """
    """
    A = A.tocsr()
    A_indices = A.indices
    A_indptr = A.indptr
    A_data = A.data
    # B_data = np.array(B.flatten()).squeeze()

    # estimate the number of non-zero triplets
    # could be passed as an argument
    non_zero_count_A = A_indptr[1:] - A_indptr[:-1]
    #    i, _ = B.nonzero()
    #    non_zero_count_B = np.bincount(i, minlength=B.shape[0])
    estimated_result_size = (non_zero_count_A * B.shape[1]).sum()

    rows = np.zeros(estimated_result_size).astype(np.int32)
    cols = np.zeros(estimated_result_size).astype(np.int32)
    vals = np.zeros(estimated_result_size).astype(np.float64)

    n = lib.kron_product_by_line_2(A.shape[0], A.shape[1], B.shape[1],
                                   A_indices, A_indptr, A_data,
                                   B,
                                   rows, cols, vals)
    # print(n, estimated_result_size, n/estimated_result_size)

    return scipy.sparse.coo_matrix((vals[:n], (rows[:n], cols[:n])),
                                   shape=(A.shape[0], A.shape[1] * B.shape[1]))



class CooMatrixBuff:

    def __init__(self, shape, estimated_nnz, increment=1.3):
        self.shape = shape
        self.size = estimated_nnz
        self.increment = increment
        self.i = np.zeros(self.size).astype(np.int64)
        self.j = np.zeros(self.size).astype(np.int64)
        self.val = np.zeros(self.size)
        self.ptr = 0

    def _resize(self):
        logging.warning('need to resize CooMatrixBuff: revise your estimates of non-zero terms !')
        new_size = self.size * self.increment
        self.i = np.resize(self.i, new_size)
        self.j = np.resize(self.j, new_size)
        self.val = np.resize(self.val, new_size)
        self.i[self.ptr:] = 0
        self.j[self.ptr:] = 0
        self.val[self.ptr:] = 0.
        self.size = new_size

    def append(self, i, j, val, free_pars_only=False):
        """
        """
        logging.info(f'appending {len(i)} to buffer at location {self.ptr}')
        sz = len(i)
        assert (len(j) == sz) and (len(val) == sz)

        if (self.ptr + sz) > self.size:
            self._resize()

        # if free_pars_only:
        #     sz = lib.append(len(i), self.ptr,
        #                     i.astype(np.int64), j.astype(np.int64), val,
        #                     self.i, self.j, self.val,
        #                     1)
        #     self.ptr += sz
        # else:
        #     sz = lib.append(len(i), self.ptr,
        #                     i.astype(np.int64), j.astype(np.int64), val,
        #                     self.i, self.j, self.val,
        #                     0)
        #     self.ptr += sz

        if free_pars_only:
            idx = j >= 0
            sz = idx.sum()
            self.i[self.ptr:self.ptr+sz] = i[idx]
            self.j[self.ptr:self.ptr+sz] = j[idx]
            self.val[self.ptr:self.ptr+sz] = val[idx]
            self.ptr += sz
        else:
            self.i[self.ptr:self.ptr+sz] = i
            self.j[self.ptr:self.ptr+sz] = j
            self.val[self.ptr:self.ptr+sz] = val
            self.ptr += sz
        logging.info('done')

    def tocoo(self):
        r = scipy.sparse.coo_matrix((self.val[:self.ptr],
                                     (self.i[:self.ptr], self.j[:self.ptr])),
                                    self.shape)
        return r


class CooMatrixBuff2:

    def __init__(self, shape, increment=1.3):
        self.shape = shape
        self.increment = increment
        self._row = []
        self._col = []
        self._data = []
        self._idx = []

    def append(self, i, j, val):
        """
        """
        self._row.append(i)
        self._col.append(j)
        self._data.append(val)
        self._idx.append(j>=0)

    def tocoo(self):
        idx = np.hstack(self._idx)
        r = scipy.sparse.coo_matrix((np.hstack(self._data)[idx],
                                     (np.hstack(self._row)[idx],
                                      np.hstack(self._col)[idx])),
                                    self.shape)
        return r

    @property
    def row(self):
        return self._row

    @property
    def col(self):
        return self._col

    @property
    def data(self):
        return self._data





import numpy as np
from scipy import sparse



class Regularization:
    """Regularization penalty

    TODO: it would be a good idea if the adaptive regularization was performed
    here. Or at least if there was a method to do it on the fly here.

    .. note:: in order to save time, the regularization matrix is precomputed.
              This means that whenever the structure of the parameter vector is
              modified (i.e. blocks are added or removed), it should be
              recomputed. Note that it does not apply to the fact that we fix or
              free parameters.

    """
    def __init__(self, block_name='theta', mu0=2., mu2=1.):
        """
        """
        self.block_name = block_name
        self.mu0 = mu0
        self.mu2 = mu2
        self.matrix = None
        self.reg_lambda = None # adaptive regularization

    def init_pars(self, pars):
        """
        """
        self.matrix = self.init_reg_matrix(pars)

    def init_reg_matrix(self, pars):
        """precompute the full regularization  matrix
        """
        # reset reg_matrix and adaptive regularization
        self.matrix = None#
        self.reg_lambda = None

        pars_full = pars.copy()
        pars_full.release()

        n = len(pars_full[self.block_name].full)
        N = len(pars_full.full)

        # order 0
        P = sparse.dia_matrix((np.ones(N), [0]), shape=(n, n)).tocoo()
        i = pars_full.indexof(P.row)
        j = pars_full.indexof(P.col)
        P = sparse.coo_matrix((P.data, (i,j)), shape=(N,N))

        # second order penality, from the laplacian matrix and slightly hacked
        # to make the end of the diagonal similar to the beginning
        data = np.ones((n,3))
        data[:,1] = -2
        Q = sparse.dia_matrix((data.T, [0,1,2]), shape=[n,n]).tocoo()
        # print(Q.todense())
        i = pars_full.indexof(Q.row)
        j = pars_full.indexof(Q.col)
        Q = sparse.coo_matrix((Q.data, (i,j)), shape=(N,N))
        A = (Q.T @ Q).tolil()
        B = A[:3,:3]
        A[-3:,-3:] = B[::-1,::-1]

        return self.mu0 * P + self.mu2 * A

    def adapt_regularization_strength(self, J_model, mu=1., mu_model=1.E-6):
        """
        """
        J = J_model
        cols = np.arange(J.shape[1])
        bins = np.arange(-0.5, J.shape[1]+0.5, 1.)
        d = np.digitize(J.col, bins, right=False) - 1
        l = np.array([0. if (d==col).sum() == 0 else J.data[d==col].max() for col in cols])
        idx = l<0.5
        l[idx] = mu
        l[~idx] = mu_model
        self.reg_lambda = l

        L = sparse.dia_matrix((l, [0]), shape=self.matrix.shape)
        self.matrix = self.matrix @ L

        #        import matplotlib.pyplot as plt
        #        plt.figure()
        #        plt.plot(l, 'k.')


    def __call__(self, pars, deriv=False):
        """
        """
        if self.matrix is None:
            self.matrix = self.init_reg_matrix(pars)

        pp = pars.full
        penalty = pp.T @ self.matrix @ pp

        if not deriv:
            return penalty

        n = self.matrix.shape[0]
        idx = pars.indexof(np.arange(n)) >= 0

        grad = 2. * (self.matrix @ pp)[idx]
        hess = 2. * self.matrix[:,idx][idx,:]

        return penalty, grad, hess



import numpy as np
import matplotlib.pyplot as pl
pl.ion()
from astropy.io import fits as pf
from scipy.sparse import coo_matrix, dia_matrix, dok_matrix, diags
from scipy.optimize import leastsq, fmin_ncg, check_grad, approx_fprime
from saunerie.lsqfit import Chi2
from saunerie.fitparameters import FitParameters
from saunerie.bspline import BSpline, CardinalBSplineC, integ
from croaks import NTuple
from numpy.polynomial.legendre import leggauss
from sksparse.cholmod import cholesky, cholesky_AAt
from saunerie import optim, plottools
from scipy.integrate import quad
import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.DEBUG)


from astropy.modeling import models, fitting
from scipy.integrate import simps
from astroquery.simbad import Simbad
import scipy 


# first model :                                                                                                                                                                          
# $$ y_{sn,t_{mjd}} = f_{sn} \times S\left(\frac{t_{mjd} - t_{max}}{1+s_{sn}}; \theta_{splines} \right) $$                                                                               

theta_range = [-35, 35]

customSimbad = Simbad()
customSimbad.remove_votable_fields('coordinates')
customSimbad.add_votable_fields('sptype')


class SModel(object):
    """
    Model of Type Ia supernova light curve :
    $$ y_{sn,t_{mjd}} = f_{sn} \times S\left(\frac{t_{mjd} - t_{max}}{1+s_{sn}}; \theta_{splines} \right) $$
    """
    def __init__(self, xdata, rng_spline = theta_range, grid_size = 10):
        self.nb_sn = len(np.unique(xdata['id']))
        self.spline_basis = CardinalBSplineC(grid_size, rng_spline)
        self.xdata = xdata
        self.n_theta = grid_size
        self.pars = self.init_pars(xdata) #self.para_init(xdata)

    def init_pars(self, xdata):
        p = [("flux", self.nb_sn), ("tmax", self.nb_sn), ("s", self.nb_sn), ("theta", len(self.spline_basis))]
        p = FitParameters(p)
        p['flux'].full[:] = [100.] * self.nb_sn
        p['s'].full[:] = [0.] * self.nb_sn
        #p['flux'].fix(0, val=100.) 
        p['s'].fix(0, val=0.)
        
        for i in range(self.nb_sn):
            idx = xdata['id'] == i
            tmax = (xdata[idx]['Date'] * xdata[idx]['Flux']).sum() / xdata[idx]['Flux'].sum()
            tmax = xdata[idx]['Date'].sum() / idx.sum()
            p['tmax'].full[i] = tmax 
            p['flux'].full[i] = xdata[idx]['Flux'][np.where(xdata[idx]['Flux'] == xdata[idx]['Flux'].max())]#
            if i ==0:
                p['tmax'].fix(0)
                p['flux'].fix(0)
        
        p['theta'].full[:] = self.theta_init(xdata)
        return p
    
    def theta_init(self, xdata, plot = False):
        #dat = generate_Lc(1)
        data = xdata[xdata['id'] == 0]
        #phase = dat['Date'] - dat['Date'][dat['Flux'] == np.unique(dat['Flux']).max()]
        phase = data['Date'] - data['Date'][data['Flux'] == np.unique(data['Flux']).max()]
        y = data['Flux']/max(data['Flux'])
        jacobian = self.spline_basis.eval(phase)
        H = jacobian.T * jacobian + 0.01 * diags(np.ones(jacobian.shape[1]))
        factor = cholesky(H.tocsc())
        if plot:
            pl.figure()
            pl.plot(data['Date'], y, 'b.')
            pl.plot(data['Date'], jacobian*factor(jacobian.T * y), 'k.')
            pl.show()
        return factor(jacobian.T * y)
        
        
    def __call__(self, p, jac=False, xdata=None):
        if xdata is None:
            xdata = self.xdata
        
        self.pars.free = p
        ide = xdata['id']
        sp1 = (1. + self.pars['s'].full[ide])
        phases = (xdata['Date'] - self.pars['tmax'].full[ide]) / sp1 #(sp1* (1 + xdata['ZP']))

        flx = self.pars['flux'].full[ide]
        jacobian = self.spline_basis.eval(phases)
        v = flx * (J * self.pars['theta'].full)

        if jac:
            N = len(xdata)
            n = len(self.spline_basis)
            isn = np.arange(N)

            inter_t = xdata[xdata['id']==0].shape[0]
            dB = self.spline_basis.deriv(phases)

            # derivée flux:
            ii_f = isn
            jj_f = self.pars['flux'].indexof(ide)
            val_f = J * self.pars['theta'].full

            #derivee tmax:
            ii_t = isn
            jj_t = self.pars['tmax'].indexof(ide)
            val_t = self.pars['flux'].full[ide] * (-1/sp1) * (dB * self.pars['theta'].full) #(-1/(sp1 * (1 + xdata['ZP']))) * (dB * self.pars['theta'].full)

            #derivee s:
            ii_s = isn
            jj_s = self.pars['s'].indexof(ide)
            val_s = self.pars['flux'].full[ide] * (-phases/sp1) * (dB * self.pars['theta'].full)

            #derivee theta:
            ii_theta = J.row
            jj_theta = self.pars['theta'].indexof(J.col)
            val_theta = self.pars['flux'].full[ide[J.row]] * J.data

            i = np.hstack((ii_f, ii_t, ii_s, ii_theta))
            j = np.hstack((jj_f, jj_t, jj_s, jj_theta))
            val_jac = np.hstack((val_f, val_t, val_s, val_theta))

            idx_fix = np.where(j != -1)
            Jac = coo_matrix((val_jac[idx_fix], (i[idx_fix],j[idx_fix])), shape = (N, len(self.pars.free)))
                                                                                                                                                                                         
            return  v , Jac
        else :
            return v


class VModel(object):
    """
    Two variance model can be implemented.
    variance model:  - variance = v **2 * flux**2, here v is a parameter and flux is the measure of the SN light curve 
                     - variance = v **2 where the variance does not depend of the SN flux
    
    There are two models, labelled by the input "lab", 'cst' stands for constant and refers to the second one
    and anything else stands for 'dependant' and refers to the first one. 
    """
    def __init__(self, n_theta, data, lab = 'cst'):
        """
        Initialisation of the model.
        Input : n_theta = Int, is the number of spline use in the model of light curves
                data = NTuple, is the light curve data   
                lab = str, is the type of variance model wanted
        Output : Varience model 
        """
        #self.value = model_sn
        self.data = data
        self.nb_sn = len(np.unique(data['id']))
        self.nb_theta = n_theta
        self.pars = self.para_init()
        self.lab = lab

    def para_init(self):
        """
        Initialisation of the parameters of the model.
        the parameter v represent the standard deviation : sigma = np.sqrt(variance), one per Supernova

        Input :  

        Output : FitParameters, class of parameter
        """
        fp = [("v", self.nb_sn)]
        fp = FitParameters(fp)
        fp['v'] = 0.1 #* 50**2 #* np.arange(self.nb_sn) + np.arange(self.nb_sn)**2
        return fp

    def __call__(self, value, p, J=None):
        """
        Return the evalutaion of the model given a set of parameter p.

        Input : - value = NTuple, is the value of the light curve model 
                - p = Array, is the VARIANCE MODEL parameters
                - J = Array, is the matrix of the derivation of the LIGHT CURVE MODEL.

        Output : - if J is given : Array, a diagonal matrix of the values of the variance model evaluate with the given parameter. 
                 - else : - Array the same matrix as describe above 
                          - dV = coo_matrix , sparse matrix of the derivative of the VARIANCE MODEL.
        """
        ide = self.data['id']
        if J == None:
            if self.lab == 'cst':
                return np.diag(p[ide]**2) #np.diag(p[ide]) 
            else :
                return np.diag((p[ide]*(value))**2) #np.diag(p[ide]*(value)**2)
        else :
            N = self.nb_theta + 4 * self.nb_sn
            n = len(value)
            isn = np.arange(n)

            # derive parametre de J:
            ii_J = J.col
            jj_J = J.row
            if self.lab == 'cst':
                val_J = np.zeros((len(J.data))) #2 * value[ide[J.row]] * J.data * p[ide[J.row]]
            else :
                val_J = 2 * value[ide[J.row]] * J.data * p[ide[J.row]]**2

            # derive V_{n}:
            ii_v = J.shape[1] + ide
            jj_v = isn
            if self.lab == 'cst':
                val_v = np.ones(len(value)) #value**2
            else :
                val_v = 2 * p[ide] * value**2
                
            i = np.hstack((ii_J, ii_v))
            j = np.hstack((jj_J, jj_v))
            val_jac = np.hstack((val_J, val_v))

            dV = coo_matrix((val_jac, (i,j)), shape = (N, n))
            if self.lab	== 'cst':
                return np.diag(p[ide]**2), dV #*(value)**2), dV
            else :
                return np.diag((p[ide]*(value))**2), dV


def generate_Lc(N, sigma_err, variance_sn, rng = [-70, 180], plot=False, lab = 'cst'):
    """
    Creation of SN light curves as gaussian. Here each tmax, stetch and maximum of the flux are the same for all SNe, but variation can be uncomment.

    Input : - N = int, is the number of SN light curves wanted.
            - sigma_err = float, is the measure error wanted on each point (can be change with a standard deviation if needed).
            - variance_sn = float, is the variance wanted on each SN, that we want to model with the Variance model.
            - rng = list , is the range of time, expressed in days
            - plot = bool, to plot the light curves
            - lab = str, to have a 'cst' (constant) variance for all SN, or a variance dependent of the flux. 

    Output : - Data = NTuple, is the new light curves.
    """
    x,y,yerr = [], [], []
    ide = []
    for i in range(N):
        tmax = 0. #np.random.uniform(theta_range[0]+15, theta_range[1]-15) #rng[0], rng[1]) #0.
        sig = 5. #np.random.uniform(1., 10.) #5.
        norm = 100. #np.random.uniform(50., 1000.) #100.

        npts = np.random.randint(5,8)*2+1
        
        xx = np.linspace(-20, 20, npts) #np.random.uniform(-3*sig, 3*sig, size=npts)
        yy = norm * np.exp(-0.5 * (xx/sig)**2)
        xx += tmax
        yy_err = sigma_err #np.random.normal(scale= sigma_err, size=len(yy))
        if lab == 'cst':
            yy += variance_sn #np.random.normal(scale=np.sqrt(0.1), size=len(yy)) #* yy**2) , size=len(yy)) 
        else :
            yy += np.random.normal(scale=np.sqrt((0.1 * yy)**2) , size=len(yy)) 
        x.append(xx)
        y.append(yy)
        ide.append(np.full(len(xx), i))
        yerr.append(np.full(len(xx), yy_err))

    x = np.hstack(x)

    data = NTuple(len(x),dtype=[('Date', float), ('Flux', float), ('FluxErr', float),('id',int)])
    data['Date'] = np.hstack(x)
    data['Flux'] = np.hstack(y)
    data['FluxErr'] = np.hstack(yerr)
    data['id'] = np.hstack(ide)
    
    if plot :
        for i in np.unique(data['id']):
            idx = data['id'] == i
            tmax = data['Flux'][idx].max()
            phase = data['Date'][idx]
            pl.errorbar(phase, data['Flux'][idx], yerr=data['FluxErr'][idx], fmt='.')

    return data



def hessian(x, w, p, g, J, dV, reml):
    """
    Create Hessian matrix needed to minimized the chi2, in order to fit models of light curves and variances on data, using Newton-Raphson's method.
    Appearing on the left hand side of the equation.
 
    Input : - x = Array, is the 'Date' of the given light curve measurements.
            - w = Array, is the matrix in which each diagonal element is the inverse variance of the measurement ( size : len(x) * len(x)).
            - p = Array, is the parameter of the LIGTH CURVE model. 
            - g = Array, is the parameter of the VARIANCE model.
            - J = coo_matrix, is the Jacobian of the Light curev model.
            - dV = coo_matrix, is the Jacobian of the variance model.
            - reml = bool, if the reml technic is needed
    
    Output : - H = coo_matrix, sparse hessian matrix
    """
    H0 = 2 * J.T * w * J
    H = coo_matrix(H0)
    H00 = H.data
    i00 = H.row
    j00 = H.col
    
    def hess_reml(n, m, x = x, dV = dV.toarray(), w = w):
        """
        calculate the reml additional term the appears in the minimization of the chi2.
        NOT use for now.
        """
        X = np.diag(x+1)
        kl = np.linalg.inv(np.linalg.multi_dot((X.T, w.toarray(), X)))
        inte = np.linalg.multi_dot((np.diag(dV[n]), w.toarray(), X, kl, X.T, w.toarray(), np.diag(dV[m]))) + np.linalg.multi_dot((np.diag(dV[n]), w.toarray(), np.diag(dV[m]))) + np.linalg.multi_dot((np.diag(dV[m]), w.toarray(), np.diag(dV[n])))
        hess = np.linalg.multi_dot((kl, X.T, w.toarray(), inte, w.toarray(), X))
        return hess.trace()
    
    def trace(i, j, w = w, dV = dV, reml = reml):
        """
        Calculate the trace term of the product of 2 derivative (wrt the parameters i and j)  appearing in the equation.
        
        Input : - i = int, parameters of which we want to take de derivative.
                - j = int, parameters of which we want to take de derivative.
                - w = Array, is the matrix in which each diagonal element is the inverse variance of the measurement ( size : len(data) * len(data)).
                - dV = coo_matrix, is the Jacobian of the variance model.
                - reml = bool, if the reml technic is needed 
        Output : - float, result of the trace.
        """
        dV = dV.toarray()
        dVi = coo_matrix(np.diag(dV[i]))
        dVj = coo_matrix(np.diag(dV[j]))
        aj = 0
        if reml :
            aj = hess_reml(i, j)
        return (w * dVi * w * dVj).toarray().trace() + aj

    npar = p.shape[0] + g.shape[0]
    Tr = np.array([trace(i,j) for i in range(npar) for j in range(i, npar)])
    ii = [i for i in range(npar) for j in range(i, npar)]
    jj = [j for i in range(npar) for j in range(i, npar)]
    
    H_l = coo_matrix((np.hstack((Tr,Tr)), (np.hstack((ii,jj)), np.hstack((jj, ii)))), shape = ( npar, npar))
    H_l = coo_matrix((np.hstack((H_l.data, -np.diag(H_l.toarray())/2.)), (np.hstack((H_l.row, range(H_l.toarray().shape[0]))), np.hstack((H_l.col, range(H_l.toarray().shape[0]))))), shape = (npar,npar))
    H = coo_matrix((H00, (i00, j00)), shape = ( npar, npar))
    return H + H_l

def grad(res, x, w, J, dV, p, g, reml):
    """
    Create gradient vector needed to minimized the chi2, in order to fit models of light curves and variances on data, using Newton-Raphson's method.
    Appearing on the right hand side of the equation.  
    
    Input : - res = Array, is the residual of the measurement substracted by the model.
            - x = Array, is the 'Date' of the given light curve measurements.
            - w = Array, is the matrix in which each diagonal element is the inverse variance of the measurement ( size : len(x) * len(x)).
            - J = coo_matrix, is the Jacobian of the Light curev model.
            - dV = coo_matrix, is the Jacobian of the variance model.
            - p = Array, is the parameter of the LIGTH CURVE model.
            - g = Array, is the parameter of the VARIANCE model.
            - reml = bool, if the reml technic is needed.

    Output : - (-1) * resul = Array, the needed gradient vector. 
    """
    def grad_reml(n, x = x, dV = dV.toarray(), w = w):
        """
        Calculate the reml additional term the appears in the minimization of the chi2.
        NOT use for now.
        """
        X = np.diag(x+1)
        kl = np.linalg.inv(np.linalg.multi_dot((X.T, w.toarray(), X)))
        grad = np.linalg.multi_dot((kl, X.T, w.toarray(), np.diag(dV[n]), w.toarray(), X)).trace()
        return - grad

    def grad_theta(i, J = J, res = res, dV = dV, w = w):
        """
        Calculate the term i of the gradient vector.
        
        Input : - i = int, represent the i-th parameter term of the gradient.
  
        Output : - Float, the ith term 
        """
        aj = 0
        if reml :
            aj = grad_reml(i)
        grad = (w.toarray().dot(np.diag(dV.toarray()[i]))).trace() - np.linalg.multi_dot([res.T, w.toarray(), np.diag(dV.toarray()[i]), w.toarray(), res])
        if i < p.shape[0]:
            grad += - 2 * np.linalg.multi_dot([J.toarray().T[i], w.toarray(), res])
        return grad + aj
    resul = np.array([grad_theta(i) for i in range(p.shape[0] + g.shape[0])])
    return -resul

def plot_cl(data, model, p, g, chi2= 'before', lab = 'cst'):
    """
    Plot the light curves data points and the fitted models (of the light curves and the variance).
    
    Input : - data = NTuple, light cruve data.
            - model = SModel, Light curve model.
            - p = Array, light curve model parameters.
            - g = Array, variance model parameters.
            - chi2 = str, 'before' of 'after' the minimization because plots are differents.
            - lab = str, 'cst' or 'dep' labeled the variance dependence on the flux.
    
    Output : the plot of SN light curves (data point and the fit)  with variances (real dependence and the fit one). 
    
    """
    if chi2 =='before':
        fig = pl.figure()
        ax1 = pl.subplot(211)
        ax2 = pl.subplot(212, sharex = ax1)
        phase = data['Date']
        ax1.errorbar(phase, data['Flux'], yerr = data['FluxErr'], marker='.', ls='', color='k')
        for i in np.unique(data['id']):
            d = data[data['id'] == i]
            date = np.linspace(d['Date'].min(), d['Date'].max())
            u = np.full(len(date), i).astype(int)
            d_sn = np.rec.fromarrays((date, u), names=['Date','id'])
            v = model(p, jac=0, xdata=d_sn)
            ax1.plot(d_sn['Date'], v, color=pl.cm.jet(i), ls='-')
        v = model(p, jac=0)
        ax2.plot(phase, data['Flux']-v,'k.')

    
    elif chi2 == 'after':
        fig = pl.figure()
        ax1 = pl.subplot(211)
        ax2 = pl.subplot(212, sharex = ax1)
        phase = data['Date']
        #ax1.errorbar(phase, data['Flux'], yerr = data['FluxErr'], marker='.', ls='', color='k')                                                                                                                       
        for i in np.unique(data['id']):
            d = data[data['id'] == i]
            date = np.linspace(d['Date'].min(), d['Date'].max())
            u = np.full(len(date), i).astype(int)
            err = np.full(len(date), d['FluxErr'][0]).astype(float)
            d_sn = np.rec.fromarrays((date,err,u), names=['Date', 'FluxErr','id'])
            v = model(p, jac=0, xdata=d_sn)
            ax1.errorbar(d['Date'], d['Flux'], yerr = d['FluxErr'], marker='.', ls='', color=pl.cm.jet(i**2 + 10 *i))
            ax1.plot(d_sn['Date'], v, color=pl.cm.jet(i**2 + 10 *i), ls='-', label = str(i), linewidth = 0.5)
            if lab == 'cst':
                ax1.plot(d_sn['Date'], v + 3*(np.sqrt(g[i]**2 + err)), color=pl.cm.jet(i**2 + 10 *i), ls='--', linewidth = 0.5)
                ax1.plot(d_sn['Date'], v - 3*(np.sqrt(g[i]**2 + err)), color=pl.cm.jet(i**2 + 10 *i), ls='--', linewidth = 0.5)
            else : 
                ax1.plot(d_sn['Date'], v + 3*(np.sqrt((g[i] * v)**2 + err)), color=pl.cm.jet(i**2 + 10 *i), ls='--', linewidth = 0.5)
                ax1.plot(d_sn['Date'], v - 3*(np.sqrt((g[i] * v)**2 + err)), color=pl.cm.jet(i**2 + 10 *i), ls='--', linewidth = 0.5)
        ax1.legend()
        v = model(p, jac=0)
        ax2.plot(phase, data['Flux']-v,'k.')
    

#data = generate_Lc(N = 100, sigma_err = 2, npts = 11, lab = lab)
#def fit_sqp_lin(N, sigma_err, max_iter = 100, ll = 0.1, dchi2_min = 0.1, reml = False, lab = 'cst'):
def fit_sqp_lin(data, max_iter = 10, ll = 0.1, dchi2_min = 0.1, reml = False, lab = 'cst', plot = True):
    """
    Minimization of the chi2.
    
    Input : - data = NTuple(['Date', 'Flux', 'FluxErr']), light curves data (if generated with the function above generate_LC, the format is ok).
            - max_iter = int, is the maximum number of iteration wanted.
            - ll = float, help to inverse the hessian matrix, by adding somthing one its diagonal.
            - dchi2_min = float, variation of chi2 desired to reach.
            - reml = bool, to consider the reml technic, NOT use here !
            - lab = str, 'cst' or 'dep' labbeled the variance model dependance on the light curve flux.
            - plot = bool, True will plot SN light curves given by the fonction plot_cl.
    
    Output : - p = Array, the final parameter of the light curve model.
             - g  = Array, the final parameter of the variance model.
    """
    #data = generate_Lc(N = N, sigma_err = sigma_err, npts = 11, lab = lab)
    model = SModel(data)
    var_model = VModel(model.n_theta, data, lab = lab)
    y = data['Flux']
    x = data['Date']
    sig = data['FluxErr']
    
    p = model.pars.free
    g = var_model.pars.free
    dp = np.zeros(p.shape)
    dg = np.zeros(g.shape)

    v, J = model(p, jac = True)
    V, dV = var_model(v, g, J=J)

    variance = np.diag(V) + sig**2
    w = np.diag(1/variance)

    res = y - v
    chi2 = (np.linalg.multi_dot([res.T,w,res]) + len(res) * np.log(np.pi) + np.log(np.diag(w)).sum())/ (len(data) - (p.shape[0] + g.shape[0]))
    Dchi = 0.
    niter = 0

    #plot_cl(data = data, model = model, p = p, g = g, chi2= 'before', lab = lab)
    
    while niter < max_iter:
        v, J = model(p, jac = True)
        V, dV = var_model(v, g, J=J)

        variance = np.diag(V) + sig**2
        w = coo_matrix(np.diag(1/variance))
        
        res = y - v
        H = hessian(x = x, w = w, p = p, g = g, J = J, dV = dV, reml = reml) + ll * diags(np.ones(p.shape[0]+ g.shape[0]))
        resul = grad(res = res, x = x, w = w, J = J, dV = dV, p = p, g = g, reml = reml)

        var = np.linalg.inv(H.toarray()).dot(resul)
        dp = var[:p.shape[0]]
        dg = var[p.shape[0]:]
            
        variance = np.diag(V) + sig**2
        w = np.diag(1/variance)

        # here are the two ways of implementing the line search : the first one, only on the light cures parameters & the second on both models parameters.
        # One can comment or decomment the wanted fonction phi and crit.
        
        """
        def phi(p):
            return ((w * ( y - model(p, jac = False)))**2).sum()

        def crit(t):
            return phi(p + t * dp) 
        """
        vect = np.hstack((p,g))

        def phi(vect, n = len(p)):
            p = vect[:n]
            g = vect[n:]
            val = model(p)
            V = var_model(val, g)
            variance = np.diag(V) + sig**2
            w = coo_matrix(np.diag(1/variance))
            return ((w * ( y - val))**2).sum()

        def crit(t):
            return phi(vect + t * var)
        
        
        t, fval, ni, funcalls = scipy.optimize.brent(crit, brack=(0,1), full_output=True)
        
        p = p + t * dp
        g = g + t * dg

        # if the first line search is choosen, there are the corresponding parameter variations. 
        #p = p + dp                                                                                                                                                                                            
        #g = g + dg

        print('p', p, '\n', 'dp', dp, '\n', 't', t, '\n', 'g', g, '\n', 'dg', dg,'\n')    
        v = model(p, jac=False)                                                                                                                                                                                  
        V = var_model(v, g)                                                                                                                       
        res = y - v

        variance = np.diag(V) + sig**2
        w = np.diag(1/variance)

        new_chi2 = (np.linalg.multi_dot([res.T,w,res]) + len(res) * np.log(np.pi) + np.log(np.diag(w)).sum())/ (len(data) - (p.shape[0] + g.shape[0]))
        dchi2 = new_chi2 - chi2
        logging.info('iter %f : chi2: %f -> %f [dchi2=%f]' % (niter, chi2, new_chi2, dchi2))
    
        chi2 = new_chi2
        if niter >= 1 :
            if dchi2 > 0.:
                logging.info('increasing chi2 !')
                niter += 1
                continue

            if abs(dchi2) <= dchi2_min:
                logging.info('dchi2 < dchi2_min, exiting')
                niter += 1
                break

            if niter > max_iter:
                break
    
        niter += 1
    if plot :
        plot_cl(data = data, model = model, p = p, g = g, chi2= 'after', lab = lab)
    return p, g
    
def control(real = 10, N = 10, sigma_err = 2., variance_sn = 0.01, lab = 'cst', nbins = 25, plot = False):
    """
    Control plot : histogram of the variance parameters for found for all SNe.

    Input : - real = int, number of realisation of the fit.
            - N = int, number of SNe fitting in each "real" realisation.
            - sigma_err = float, measure error we want on the SNe light curves
            - variance_sn = float, variance we want to be modeled by the variance model.
            - lab = str 'cst' or 'dep', type of variance model we want to use.
            - nbins = int, number of bins for the histogram.
            - plot = bool, True if we want to see each Sne light curve plot with the foncion plot_cl.
    
    Output : histogram
    """
    g = []
    for i in range(real):
        data = generate_Lc(N = N, sigma_err = sigma_err, variance_sn = variance_sn, lab = lab)
        p0, g0 = fit_sqp_lin(data, max_iter = 20, lab = lab, plot = plot)
        g.append(g0)

    g = np.array(g)
    pl.figure()
    xh, yh, hh = pl.hist(np.abs(g.ravel()), bins=nbins)#, range = (,0.01))

    med = np.median(g)
    moy = g.mean()

    pl.plot(variance_sn * np.ones(10), np.linspace(0,xh.max(),10), 'r--')
    pl.plot(moy * np.ones(10), np.linspace(0,xh.max(),10), 'k--')
    pl.plot(med * np.ones(10), np.linspace(0,xh.max(),10), 'b--')

    pl.title('%i real  de %i SN : v = %f ( => v_simu moy : %f med : %f)' % (real, N, variance_sn, moy, med))


def plot_regimes(var = [0.001, 0.01, 0.1, 1., 10,], lab = 'dep', real = 10, N = 10):
    """
    plot histogram for the var differents value of the variance_sn of the control function.

    Input : - var, list of float, the variance wanted for the variance_sn of the function control.
            - lab = 'cst' or 'dep', the variance model depnedance on the flux
            - real = int, number of realisation of the fit.
            - N = int, number of SNe fitting in each "real" realisation.
    """
    for i in var :
        control(variance_sn = i, lab = lab, real = real, N = N)
        
    



"""SN light curve eval unit
"""

import logging

import time
import numpy as np
import scipy

import numexpr as ne

try:
    from sparse_dot_mkl import gram_matrix_mkl, dot_product_mkl
except:
    logging.warning('module: `sparse_dot_mkl` not available')
else:
    logging.info('sparse_dot_mkl found. Building hessian should be faster.')

from ...sparseutils import kron_product_by_line, CooMatrixBuff, CooMatrixBuff2
from bbf import magsys

class LightcurveEvalUnitTz(object):
    r"""Evaluate the light curve of a single SN in a single band

    Given a set of dates, and the :math:`(X_0, X_1, c, t_{max})` parameters of
    the SN, compute the quantities:

    .. math::

         X_0 \times (1+z) \times \left[I_0 + X_1 I_1\right]

    with

    .. math::
         I_{0|1} = \int M_{0|1}(\lambda, \mathrm{p}) \frac{\lambda}{hc} T((1+z)\lambda) 10^{0.4\ c\ P(\lambda)} d\lambda

    In practice, we can speed up the computations by decomposing the filter on a
    spline basis: :math:`T(\lambda (1+z)) = \sum_q t_q(z) B_q(\lambda)= \sum_q
    (1+z) t_q B_q(\lambda)`, which allows to decompose the integral above as:

    .. math::
        I_{0|1} = \sum_{k\ell q} \theta_{0|1, k\ell} t_q(z) B_k(\mathrm{p})\ \int B_\ell(\lambda)
        \frac{\lambda}{hc} B_q(\lambda)\ 10^{0.4\ c\ P(\lambda)} d\lambda

    We use the fact that the color law is a slowly variable function of
    :math:`\lambda` (compared to the individual splines) to expel the color law
    from the integral:

    .. math::
       \int B_\ell(\lambda) \frac{\lambda}{hc} B_q(\lambda) CL(\lambda) d\lambda \approx
       CL(\lambda^{\mathrm{eff}}_{\ell q})\ \times \int B_\ell(\lambda) \frac{\lambda}{hc} B_q(\lambda) d\lambda

    up to second order, if we define:

    .. math::
       \lambda^{\mathrm{eff}}_{\ell q} = \frac{\int B_\ell(\lambda)
       \lambda^2 B_q(\lambda) d\lambda}{\int B_\ell(\lambda) \lambda B_q(\lambda) d\lambda}


    and so, defining: :math:`\Lambda^{\mathrm{eff}} =
    (\lambda^{\mathrm{eff}}_{\ell q})` and ::math:`G = (G_{\ell q}) = (\int
    B_\ell(\lambda)\frac{\lambda}{hc} B_q(\lambda)) d\lambda`, the integral
    evaluation reduces to:

    .. math::
        I_{0|1} = J\ \cdot \Theta\ \cdot \left[G \Lambda^{\mathrm{eff}}\right] \cdot t_z

    where :math:`J`, :math:`G`, :math:`\Lambda^{\mathrm{eff}}` and :math:`t_z`
    are sparse and where :math:`G, \Lambda^{\mathrm{eff}}` can be precomputed
    once for all.

    Matrix :math:`G` is called the Gramian of the model and filter spline bases.
    It is precomputed exactly, using Gaussian quadrature when the model is
    instantiated.

    This strategy decomposes :math:`T((1+z)\lambda)` and work with a single
    Gramian and an SN-dependant :math:`t_z` vector. This option is probably
    computationally less intensive, as we can factorize the evaluation of the
    color law.

    """

    def __init__(self, lcdata, model):
        r"""Constructor

        Retrieve the band shape :math:`T(\lambda)` and project
        :math:`T(\lambda(1+z))` on the basis

        Parameters
        ----------
        lcdata : nacl.dataset.LcData
            Photometric of one Lc of one SN nacl data class.
        model : nacl.models.salt.SALT2Like
            Model.

        """
        self.lcdata = lcdata
        # self.data = lcdata.data # we no longer use this
        self.model = model
        self.z = lcdata.z[0]

        # TODO: here, we take only the default band
        # we should specialize and take the band at default location
        # tr = model.filter_db.transmission_db[self.band]
        self.band = lcdata.band[0]
        self.tqz, _ = model.filter_db.insert(self.band, z=self.z)
        # set to zero very small values
        self.tqz[(np.abs(self.tqz)/self.tqz.max() < 1e-10)] = 0.
        
        # self.sn_id = lcdata.sn_id
        self.sn_index = lcdata.sn_index[0]

        #        self.wl_basis_size = len(model.basis.bx)
        self.ph_basis_size = len(model.basis.by)

        # just in case
        self.corr = np.ones(len(lcdata.band_index))

    def __call__(self, pars, jac=False, debug_mode=False):
        """Evaluate the light curve.

        Parameters
        ----------
        jac : bool
            Whether it return the jacobian matrix
        debug_mode : bool
            if True, return the model components instead of the
            model results.

        Returns
        ----------
        val : numpy.array
            Model evaluation
        jacobian : scipy.sparse.csr_matrix, optional
            Jacobian matrix.

        """
        t0 = time.perf_counter()
        sn_index = self.sn_index
        zz = 1. + self.z
        model = self.model
        #       pars = self.model.pars

        M0 = pars['M0'].full.reshape(self.ph_basis_size, -1)
        M1 = pars['M1'].full.reshape(self.ph_basis_size, -1)
        # cl_pars = model.pars['CL'].full

        # data = self.lcdata.data
        # band_index = self.lcdata.band_index[0]

        # band_id = self.lcdata.data['band_id'][0]
        # print(self.model.bands[band_id], self.lcdata.data['Filter'][0])
        # lc_id = self.lcdata.data['lc_id'][0]
        # lc_index = self.lcdata.lc_index[0]
        sn_index = self.lcdata.sn_index[0]

        # sn-related parameters
        x0, x1 = pars['X0'].full[sn_index], pars['X1'].full[sn_index]
        c, tmax = pars['c'].full[sn_index],  pars['tmax'].full[sn_index]

        if 'eta_calib' in pars._struct.slices:
            calib_corr = 1. + pars['eta_calib'].full[self.lcdata.band_index]
        else:
            calib_corr = self.corr

        if 'kappa_color' in pars._struct.slices:
            cs_corr = 1. + pars['kappa'].full[self.lcdata.lc_index]
        else:
            cs_corr = self.corr
#        if model.calib_error_model is not None:
#            # eta = pars['eta_calib'].full[band_index]
#            calib_corr = model.calib_error_model.correction(lcdata=self.lcdata)
#        else:
#            calib_corr = np.ones(len(self.lcdata))
#        if model.color_scatter_model is not None:
#            # kappa = pars['kappa_color'].full[lc_index]
#            cs_corr = model.color_scatter_model.correction()
#        else:
#            cs_corr = 1.
        # model_to_meas_scale = self.lc_data.norm
        flux_scale = model.norm * self.lcdata.zp_scale

        phase_eval = model.phase_eval[self.lcdata.slc]
        # color law evaluation -- on the Gram \lambda_eff's
        cl = np.power(10., 0.4 * c * model.polynome_color_law.data)
        csr_color_law = scipy.sparse.csr_matrix((cl, (model.L_eff.row, model.L_eff.col)),
                                                shape=model.L_eff.shape)
        gram = model.gram.multiply(csr_color_law)
        tqz = self.tqz

        # much faster with the .dot's (instead of @)
        integral_surface_0 = phase_eval.dot(M0.dot(gram.dot(tqz)))
        integral_surface_1 = phase_eval.dot(M1.dot(gram.dot(tqz)))
        flux = flux_scale * zz * x0 * (integral_surface_0+x1*integral_surface_1) * calib_corr * cs_corr  # (1+eta)*(1+kappa)

        # if debug_mode:
        #     return model_norm * integral_surface_0, model_norm * integral_surface_1, \
        #            cl, flux, x0, x1, model_norm

        model.val.append(flux)
        if not jac:
            model.timing.append(time.perf_counter()-t0)
            return flux

        # n_data = len(self.lcdata)
        # sn_index = np.full(n_data, self.sn_index)
        # norm = self.model.norm

        # shortcut names to the internal cache holding the
        # jacobian matrix definition
        jacobian_i, jacobian_j, jacobian_val = model.jacobian_i, model.jacobian_j, model.jacobian_val

        # dMdX0
        jacobian_i.append(self.lcdata.row)
        jacobian_j.append(pars['X0'].indexof(self.lcdata.sn_index))
        assert len(self.lcdata.row) == len(pars['X0'].indexof(self.lcdata.sn_index))
        jacobian_val.append(flux_scale * zz * (integral_surface_0 + x1*integral_surface_1) * calib_corr * cs_corr)  # (1+eta) * (1+kappa))

        # dMdX1
        jacobian_i.append(self.lcdata.row)
        jacobian_j.append(pars['X1'].indexof(self.lcdata.sn_index))
        assert len(self.lcdata.row) == len(pars['X1'].indexof(self.lcdata.sn_index))
        jacobian_val.append(flux_scale * zz * x0 * integral_surface_1 * calib_corr * cs_corr)  # (1+eta) * (1+kappa))

        # dMdc
        jacobian_i.append(self.lcdata.row)
        jacobian_j.append(pars['c'].indexof(self.lcdata.sn_index))
        assert len(self.lcdata.row) == len(pars['c'].indexof(self.lcdata.sn_index))
        gram_color_law = gram.multiply(model.polynome_color_law)
        dintegral0_dc = 0.4 * np.log(10.) * phase_eval.dot(M0.dot(gram_color_law.dot(tqz)))
        dintegral1_dc = 0.4 * np.log(10.) * phase_eval.dot(M1.dot(gram_color_law.dot(tqz)))
        jacobian_val.append(flux_scale * zz * x0 * (dintegral0_dc + x1*dintegral1_dc) * calib_corr * cs_corr)  # (1+eta) * (1+kappa))

        # dMdtmax
        dphase_dtmax = model.dphase_dtmax[self.lcdata.slc]
        jacobian_i.append(self.lcdata.row)
        jacobian_j.append(pars['tmax'].indexof(self.lcdata.sn_index))
        assert len(self.lcdata.row) == len(pars['tmax'].indexof(self.lcdata.sn_index))
        dintegral0_dtmax = -dphase_dtmax.dot(M0.dot(gram.dot(tqz))) / zz
        dintegral1_dtmax = -dphase_dtmax.dot(M1.dot(gram.dot(tqz))) / zz
        jacobian_val.append(flux_scale * zz * x0 * (dintegral0_dtmax + x1*dintegral1_dtmax) * calib_corr * cs_corr)  # (1+eta)*(1+kappa))

        # dMdtheta_0
        dbase_dtheta = scipy.sparse.kron(phase_eval, gram.dot(tqz)).tocoo()
        jacobian_i.append(self.lcdata.row[dbase_dtheta.row])
        jacobian_j.append(pars['M0'].indexof(dbase_dtheta.col))
        assert len(self.lcdata.row[dbase_dtheta.row]) == len(pars['M0'].indexof(dbase_dtheta.col))
        jacobian_val.append(flux_scale[dbase_dtheta.row] * zz * x0 * dbase_dtheta.data * calib_corr[dbase_dtheta.row] * cs_corr[dbase_dtheta.row])  # (1+eta) * (1+kappa))

        # dMdtheta_1
        jacobian_i.append(self.lcdata.row[dbase_dtheta.row])
        jacobian_j.append(pars['M1'].indexof(dbase_dtheta.col))
        assert len(self.lcdata.row[dbase_dtheta.row]) == len(pars['M1'].indexof(dbase_dtheta.col))
        jacobian_val.append(flux_scale[dbase_dtheta.row] * zz * x0 * x1 * dbase_dtheta.data * calib_corr[dbase_dtheta.row] * cs_corr[dbase_dtheta.row])  # (1+eta) * (1+kappa))

        # dMdcl
        # I really don't know how to vectorize the computation
        # of these derivatives. If somebody has an idea let me know.
        buff = np.zeros(len(self.lcdata)).astype(int)
        for i, jacobian_color_law in enumerate(model.jacobian_color_law):
            buff[:] = i
            jacobian_i.append(self.lcdata.row)
            jacobian_j.append(pars['CL'].indexof(buff))
            gram_jacobian_color_law = gram.multiply(jacobian_color_law)
            jacobian_surface0_color_law = 0.4 * np.log(10.) * c * phase_eval.dot(M0.dot(
                gram_jacobian_color_law.dot(tqz)))
            jacobian_surface1_color_law = 0.4 * np.log(10.) * c * phase_eval.dot(M1.dot(
                gram_jacobian_color_law.dot(tqz)))
            jacobian_val.append(flux_scale * zz * x0 * (jacobian_surface0_color_law + x1*jacobian_surface1_color_law)
                                * calib_corr * cs_corr)  # * (1+eta) * (1+kappa))

        # dMdeta
        #        if model.calib_error_model is not None:
        if 'eta_calib' in pars._struct:
            jacobian_i.append(self.lcdata.row)
            jacobian_j.append(pars['eta_calib'].indexof(self.lcdata.band_index))
            jacobian_val.append(flux_scale * x0 * zz * (integral_surface_0 + x1*integral_surface_1) * cs_corr)  # *(1+kappa))

        # dKappa
        # if model.color_scatter_model is not None:
        if 'kappa_color' in pars._struct:
            jacobian_i.append(self.lcdata.row)
            jacobian_j.append(pars['kappa_color'].indexof(self.lcdata.band_index))
            jacobian_val.append(flux_scale * x0 * zz * (integral_surface_0 + x1*integral_surface_1) * calib_corr)  # (1+eta))

        if model.disable_cache:
            model.clear_cache()
        model.timing.append(time.perf_counter()-t0)
        return flux


class LightcurveEvalUnit:
    """An experimental eval unit for light curves

    Evaluates all light curves at once - should be faster
    """
    def __init__(self, model):
        """
        """
        self.model = model
        self.training_dataset = model.training_dataset
        self.gram = self.model.gram.todense()
        self.color_law = self.model.color_law

        nb_lightcurves = len(self.training_dataset.lc_db)
        filter_db_basis_size = len(model.filter_db.basis)
        F = np.zeros((nb_lightcurves, filter_db_basis_size))
        for lc in self.training_dataset.lc_db:
            tqz, _ = self.model.filter_db.insert(lc.band, z=lc.z)
            F[lc.lc_index, :] = tqz

        self.flux_scales = self.compute_flux_scales()
        self.flux_scales *= self.model.norm

        # filter projections
        self.filter_projections = (self.gram @ F.T).T
        self.meas_filter_projections = \
            self.filter_projections[self.training_dataset.lc_data.lc_index]


    def compute_flux_scales(self):
        """the training dataset contains fluxes and fluxerrs expressed in units defined by the observer. 
        The observer is supposed to give the connection between its unit system natural calibrated magnitudes 
        in some system (generally AB). 

        .. math::
           -2.5 \log_{10} 

        To predict these fluxes, the model values need to be multiplied by a scale, equal to:
          
        .. note::
           here, we assume that all mags are in the AB mag system. This may not be true.
        """
        zp_scale = self.training_dataset.lc_data.zp_scale
        
        # also insert the filter at z = 0, we need it
        # to compute integrals of AB spectrum
        bands = np.unique(self.training_dataset.lc_db.band)
        for b in bands:
            self.model.filter_db.insert(b, z=0.)

        ms = magsys.SNMagSys(self.model.filter_db)            
        int_AB_dict = dict(zip(bands, [10.**(0.4 * ms.get_zp(b.lower())) for b in bands]))
        int_AB = np.array([int_AB_dict[b] for b in self.training_dataset.lc_data.band])
        flux_scales = int_AB / zp_scale
        return flux_scales
        
    def __call__(self, pars, jac=False):
        """
        """
        lc_data = self.training_dataset.lc_data
        lc_db = self.training_dataset.lc_db

        wl_basis = self.model.basis.bx
        ph_basis = self.model.basis.by
        n_wl, n_ph = len(wl_basis), len(ph_basis)

        # pick matrix
        n_lc, n_meas = len(lc_db), len(lc_data)

        # phases
        zz = 1. + lc_data.z
        tmax = pars['tmax'].full[lc_data.sn_index]
        restframe_phases = (lc_data.mjd - tmax) / zz
        J_phase_sparse = ph_basis.eval(restframe_phases + self.model.delta_phase)
        J_phase = np.array(J_phase_sparse.todense())


        if 'eta_calib' in pars._struct.slices:
            calib_corr = 1. + pars['eta_calib'].full[lc_data.band_index]
        else:
            calib_corr = np.ones(len(lc_data.band_index))

        if 'kappa_color' in pars._struct.slices:
            cs_corr = 1. + pars['kappa'].full[lc_data.lc_index]
        else:
            cs_corr = np.ones(len(lc_data.lc_index))

        M0 = pars['M0'].full.reshape(n_ph, n_wl)
        M1 = pars['M1'].full.reshape(n_ph, n_wl)
        C0_ = np.array(M0.dot(self.meas_filter_projections.T))
        C0 = (J_phase * C0_.T).sum(axis=1)
        C1_ = np.array(M1.dot(self.meas_filter_projections.T))
        C1 = (J_phase * C1_.T).sum(axis=1)

        X0 = pars['X0'].full[lc_data.sn_index]
        X1 = pars['X1'].full[lc_data.sn_index]
        col = pars['c'].full[lc_data.sn_index]

        # color law - so here, we decide to move it out of the integral
        # maybe we need to add a small correction
        cl_pars = pars['CL'].full

        restframe_wavelength = lc_data.wavelength / zz
        cl_pol, J_cl_pol = self.color_law(restframe_wavelength,
                                          cl_pars, jac=jac)
        cl = np.power(10., 0.4 * col * cl_pol)

        pca = C0 + X1 * C1
        model_val = X0 * pca * cl * zz * calib_corr * cs_corr

        if not jac:
            v = np.zeros(len(self.training_dataset))
            v[lc_data.row] = self.flux_scales * model_val
            return v

        # jacobian
        N = len(self.training_dataset)
        n_free_pars = len(pars.free)

        # since the hstack is taking a lot of time and memory, we do things differently:
        # we allocate 3 large buffers for the jacobian i, j, vals, and we
        # update them in place.

        # estimated size of the derivatives
        logging.debug(' ... kron')
        K = kron_product_by_line(J_phase_sparse, self.meas_filter_projections)
        logging.debug(f'     -> done, K.nnz={K.nnz} nnz_real={(K.data != 0.).sum()} {len(K.row)}')
        estimated_size = 2 * K.nnz   # dMdM0 and dMdM1
        estimated_size += 6 * N      # local parameters (dMdX0, dMdX1, dMdcol, dMtmax, dMdeta, dMdkappa)
        nnz = len(J_cl_pol.nonzero()[0])
        estimated_size += nnz
        logging.debug(f'estimated size: {estimated_size}')

        buff = CooMatrixBuff2((N, n_free_pars)) # , estimated_size)
        self.K = K
        self.X0 = X0
        self.cl = cl
        self.calib_corr = calib_corr
        self.cs_corr = cs_corr
        self.zz = zz

        # we start with the largest derivatives: dMdM0 and dMdM1
        # dMdM0
        # we could write it as:
        # v_ = X0[K.row] * K.data * cl[K.row] * calib_corr[K.row] * cs_corr[K.row] * zz[K.row]
        # but it is slow. So, we re-arrange it as:
        i_ = lc_data.row[K.row]
        v_ = X0 * cl * calib_corr * cs_corr * zz
        # vv_, dd_ = v_[K.row], K.data
        # v_ = ne.evaluate('vv_ * dd_')
        v_ = v_[K.row] * K.data
        buff.append(i_,
                    pars['M0'].indexof(K.col),
                    v_)

        # dMdM1
        # X1_ = X1[K.row]
        buff.append(lc_data.row[K.row],
                    pars['M1'].indexof(K.col),
                    v_ * X1[K.row])

        del K
        del i_
        del v_

        # dMdtmax
        phase_basis = self.model.basis.by
        dJ = np.array(phase_basis.deriv(restframe_phases + self.model.delta_phase).todense())
        dC0 = (dJ * C0_.T).sum(axis=1)
        dC1 = (dJ * C1_.T).sum(axis=1)
        buff.append(lc_data.row,
                    pars['tmax'].indexof(lc_data.sn_index),
                    # ne.evaluate('-X0 * (dC0 + X1 * dC1) * cl * calib_corr * cs_corr'))
                    -X0 * (dC0 + X1 * dC1) * cl * calib_corr * cs_corr)

        del dJ
        del C0_
        del C1_

        # dMdcl
        JJ = scipy.sparse.coo_matrix(J_cl_pol)
        model_val_, col_, d_ = model_val[JJ.row], col[JJ.row], JJ.data
        buff.append(JJ.row,
                    pars['CL'].indexof(JJ.col),
                    # ne.evaluate('model_val_ * 0.4 * np.log(10.) * col_ * d_'))
                    # self.flux_scales[JJ.row] * model_val[JJ.row] * 0.4 * np.log(10.) * col[JJ.row] * JJ.data)
                    model_val[JJ.row] * 0.4 * np.log(10.) * col[JJ.row] * JJ.data)

        del JJ
        del model_val_
        del col_
        del d_

        # dMdX0
        buff.append(lc_data.row,
                    pars['X0'].indexof(lc_data.sn_index),
                    pca * cl * zz * calib_corr * cs_corr)

        # dMdX1
        buff.append(lc_data.row,
                    pars['X1'].indexof(lc_data.sn_index),
                    X0 * C1 * cl * zz * calib_corr * cs_corr)

        # dMdcol
        buff.append(
            lc_data.row,
            pars['c'].indexof(lc_data.sn_index),
            model_val * 0.4 * np.log(10.) * cl_pol)

        # dMdeta
        if 'eta_calib' in pars._struct:
            buff.append(lc_data.row,
                        pars['eta_calib'].indexof(lc_data.band_index),
                        X0 * pca * cl * zz * cs_corr)

        # dMkappa
        if 'kappa_color' in pars._struct:
            buff.append(lc_data.row,
                        pars['kappa_color'].indexof(lc_data.band_index),
                        X0 * pca * cl * zz * calib_corr)

        logging.debug(' -> tocoo()')
        J = buff.tocoo()
        del buff

        # multiply the data by the flux scales
        # to express fluxes in observer units
        J.data *= self.flux_scales[J.row]
        v = np.zeros(len(self.training_dataset))
        v[lc_data.row] = self.flux_scales * model_val

        return v, J

    def call_deprecated(self, pars, jac=False):
        """
        """
        lc_data = self.training_dataset.lc_data
        lc_db = self.training_dataset.lc_db

        wl_basis = self.model.basis.bx
        ph_basis = self.model.basis.by
        n_wl, n_ph = len(wl_basis), len(ph_basis)

        # pick matrix
        n_lc, n_meas = len(lc_db), len(lc_data)
        #        P = scipy.sparse.coo_matrix((np.ones(n_meas),
        #                                     (lc_data.row, lc_data.lc_index)),
        #                                    shape=(n_meas, n_lc)).tocsc()
        #        F = self.gram.dot(P.dot(self.filter_projections).T)

        # phases
        zz = 1. + lc_data.z
        tmax = pars['tmax'].full[lc_data.sn_index]
        restframe_phases = (lc_data.mjd - tmax) / zz
        J_phase_sparse = ph_basis.eval(restframe_phases + self.model.delta_phase)
        J_phase = np.array(J_phase_sparse.todense())


        if 'eta_calib' in pars._struct.slices:
            calib_corr = 1. + pars['eta_calib'].full[lc_data.band_index]
        else:
            calib_corr = np.ones(len(lc_data.band_index))

        if 'kappa_color' in pars._struct.slices:
            cs_corr = 1. + pars['kappa'].full[lc_data.lc_index]
        else:
            cs_corr = np.ones(len(lc_data.lc_index))

        M0 = pars['M0'].full.reshape(n_ph, n_wl)
        M1 = pars['M1'].full.reshape(n_ph, n_wl)
        C0_ = np.array(M0.dot(self.meas_filter_projections.T))
        C0 = (J_phase * C0_.T).sum(axis=1)
        C1_ = np.array(M1.dot(self.meas_filter_projections.T))
        C1 = (J_phase * C1_.T).sum(axis=1)

        X0 = pars['X0'].full[lc_data.sn_index]
        X1 = pars['X1'].full[lc_data.sn_index]
        col = pars['c'].full[lc_data.sn_index]

        # color law - so here, we decide to move it out of the integral
        # maybe we need to add a small correction
        cl_pars = pars['CL'].full

        restframe_wavelength = lc_data.wavelength / zz
        cl_pol, J_cl_pol = self.color_law(restframe_wavelength,
                                          cl_pars, jac=jac)
        cl = np.power(10., 0.4 * col * cl_pol)

        pca = C0 + X1 * C1
        model_val = X0 * pca * cl * zz * calib_corr * cs_corr

        if not jac:
            v = np.zeros(len(self.training_dataset))
            v[lc_data.row] = model_val
            return v


        # jacobian
        J_i, J_j, J_val = [], [], []

        N = len(self.training_dataset)
        n_free_pars = len(pars.free)
        jacobian = scipy.sparse.dok_matrix((N,n_free_pars)).tocsr()

        # dMdX0
        #        J_i.append(lc_data.row)
        #        J_j.append(pars['X0'].indexof(lc_data.sn_index))
        #        J_val.append(pca * cl * zz * calib_corr * cs_corr)
        i = lc_data.row
        j = pars['X0'].indexof(lc_data.sn_index)
        val = pca * cl * zz * calib_corr * cs_corr
        idx = j >= 0
        jacobian += scipy.sparse.coo_matrix((val[idx], (i[idx], j[idx])), shape=jacobian.shape)

        # dMdX1
        #        J_i.append(lc_data.row)
        #        J_j.append(pars['X1'].indexof(lc_data.sn_index))
        #        J_val.append(X0 * C1 * cl * zz * calib_corr * cs_corr)
        i = lc_data.row
        j = pars['X1'].indexof(lc_data.sn_index)
        val = X0 * C1 * cl * zz * calib_corr * cs_corr
        idx = j >= 0
        jacobian += scipy.sparse.coo_matrix((val[idx], (i[idx], j[idx])), shape=jacobian.shape)

        # dMdcol
        #        J_i.append(lc_data.row)
        #        J_j.append(pars['c'].indexof(lc_data.sn_index))
        #        J_val.append(model_val * 0.4 * np.log(10.) * cl_pol)
        i = lc_data.row
        j = pars['c'].indexof(lc_data.sn_index)
        val = model_val * 0.4 * np.log(10.) * cl_pol
        idx = j >= 0
        jacobian += scipy.sparse.coo_matrix((val[idx], (i[idx], j[idx])), shape=jacobian.shape)

        # dMdtmax
        phase_basis = self.model.basis.by
        dJ = np.array(phase_basis.deriv(restframe_phases + self.model.delta_phase).todense())
        # C0 = np.array(M0.dot(self.meas_filter_projections.T))
        dC0 = (dJ * C0_.T).sum(axis=1)
        # C1 = np.array(M0.dot(self.meas_filter_projections.T))
        dC1 = (dJ * C1_.T).sum(axis=1)
        #        J_i.append(lc_data.row)
        #        J_j.append(pars['tmax'].indexof(lc_data.sn_index))
        #        J_val.append(-X0 * (dC0 + X1 * dC1) * cl * calib_corr * cs_corr)
        i = lc_data.row
        j = pars['tmax'].indexof(lc_data.sn_index)
        val = -X0 * (dC0 + X1 * dC1) * cl * calib_corr * cs_corr
        idx = j >= 0
        jacobian += scipy.sparse.coo_matrix((val[idx], (i[idx], j[idx])), shape=jacobian.shape)

        # dMdM0
        # J = scipy.sparse.kron(J_phase_sparse, self.meas_filter_projections)
        # J = sparse_kron_along_cols(J_phase_sparse, self.meas_filter_projections)
        logging.debug(' ... kron')
        J = kron_product_by_line(J_phase_sparse, self.meas_filter_projections)
        logging.debug('     -> done')

        logging.debug(' ... dMdM0')
        #        J_i.append(lc_data.row[J.row])
        #        J_j.append(pars['M0'].indexof(J.col))
        #        J_val.append(X0[J.row] * J.data * cl[J.row] * calib_corr[J.row] * cs_corr[J.row] * zz[J.row])
        i = lc_data.row[J.row]
        j = pars['M0'].indexof(J.col)
        val = X0[J.row] * J.data * cl[J.row] * calib_corr[J.row] * cs_corr[J.row] * zz[J.row]
        idx = j >= 0
        jacobian += scipy.sparse.coo_matrix((val[idx], (i[idx], j[idx])), shape=jacobian.shape)
        logging.debug('done')

        # dMdM1
        logging.debug(' ... dMdM1')
        #        J_i.append(lc_data.row[J.row])
        #        J_j.append(pars['M1'].indexof(J.col))
        #        J_val.append(X0[J.row] * X1[J.row] * J.data * cl[J.row] * calib_corr[J.row] * cs_corr[J.row] * zz[J.row])
        i = lc_data.row[J.row]
        j = pars['M1'].indexof(J.col)
        val = X0[J.row] * X1[J.row] * J.data * cl[J.row] * calib_corr[J.row] * cs_corr[J.row] * zz[J.row]
        idx = j >= 0
        jacobian += scipy.sparse.coo_matrix((val[idx], (i[idx], j[idx])), shape=jacobian.shape)
        logging.debug('done')

        # saves some memory here
        del J

        # dMdcl
        logging.debug(' -> dMdcl')
        JJ = scipy.sparse.coo_matrix(J_cl_pol)
        # J_i.append(JJ.row)
        # J_j.append(pars['CL'].indexof(JJ.col))
        # J_val.append(model_val[JJ.row] * 0.4 * np.log(10.) * col[JJ.row] * JJ.data)
        i = JJ.row
        j = pars['CL'].indexof(JJ.col)
        val = model_val[JJ.row] * 0.4 * np.log(10.) * col[JJ.row] * JJ.data
        idx = j >= 0
        jacobian += scipy.sparse.coo_matrix((val[idx], (i[idx], j[idx])), shape=jacobian.shape)

        # dMdeta
        logging.debug(' -> dMdeta')
        if 'eta_calib' in pars._struct:
            #            J_i.append(lc_data.row)
            #            J_j.append(pars['eta_calib'].indexof(lc_data.band_index))
            #            J_val.append(X0 * pca * cl * zz * cs_corr)
            i = lc_data.row
            j = pars['eta_calib'].indexof(lc_data.band_index)
            val = X0 * pca * cl * zz * cs_corr
            idx = j >= 0
            jacobian += scipy.sparse.coo_matrix((val[idx], (i[idx], j[idx])), shape=jacobian.shape)

        # dMkappa
        logging.debug(' -> dMdkappa')
        if 'kappa_color' in pars._struct:
            #            J_i.append(lc_data.row)
            #            J_j.append(pars['kappa_color'].indexof(lc_data.band_index))
            #            J_val.append(X0 * pca * cl * zz * calib_corr)
            i = lc_data.row
            j = pars['kappa_color'].indexof(lc_data.band_index)
            val = X0 * pca * cl * zz * calib_corr
            idx = j >= 0
            jacobian += scipy.sparse.coo_matrix((val[idx], (i[idx], j[idx])), shape=jacobian.shape)

        # logging.info(' -> all together...')
        # J_i = np.hstack(J_i)
        # J_j = np.hstack(J_j)
        # J_val = np.hstack(J_val).astype(np.float32)
        # idx = J_j >= 0
        # N = len(self.training_dataset)
        # n_free_pars = len(pars.free)
        # JJ = scipy.sparse.coo_matrix((J_val[idx], (J_i[idx], J_j[idx])),
        #                              shape=(N, n_free_pars))

        v = np.zeros(len(self.training_dataset))
        v[lc_data.row] = model_val
        return v, jacobian


# def sparse_kron_along_cols(A, B):
#     """
#     """
#     A = A.tocsr()

#     m, n1 = A.shape
#     n2 = B.shape[1]
#     Cshape = (m, n1*n2)

#     data = []
#     col =  []
#     row =  []
#     ind2 = np.arange(n2)

#     for i in range(A.shape[0]):
#         slc1 = slice(A.indptr[i], A.indptr[i+1])
#         data1 = A.data[slc1]
#         ind1 = A.indices[slc1]
#         # slc2 = slice(B.indptr[i],B.indptr[i+1])
#         # data2 = B.data[slc2]; ind2 = B.indices[slc2]
#         data2 = np.array(B[i]).squeeze()
#         data.append(np.outer(data1, data2).ravel())
#         col.append(((ind1*n2)[:,None]+ind2).ravel())
#         row.append(np.full(len(data1) * len(data2), i))
#     data = np.concatenate(data)
#     col = np.concatenate(col)
#     row = np.concatenate(row)
#     return scipy.sparse.coo_matrix((data,(row,col)),shape=Cshape)






import numpy as np
import pylab as pl
from nacl.models.salt import SALT2Like
from nacl import dataset
from nacl.simulations.fullsim import sngen

import helpers

tds, model = helpers.generate_dataset(2, seed=42)


def test_model_normalization():
    """
    check the model normalization, i.e. check that
      - with the definition of X0 as (10Mpc)^2/d_L^2(z)
      - with the definition of the model normalization
    we can reproduce, at 0th order, the JLA light curves
    """
    # gen = sngen.DVDzSNGenerator(z_range=(0.001, 0.6))
    # sample = gen.generate_sample(nsn=10000)
    jla = dataset.read_hdf('jla.hd5')
    # zz = np.linspace(0.001, 0.6, 500)
    model = SALT2Like(jla,
                      init_from_salt2_file='salt2.npz',
                      init_from_training_dataset=True)

    return model

def test_salt2like_derivatives(block_name=None):
    """
    """
    if block_name is None:
        model.pars.release()
    else:
        model.pars.fix()
        model.pars.release(block_name)

    v, J = model(model.pars.free, jac=True)

    jac, num_der = helpers.check_grad(model, model.pars.free)

    return np.array(jac), num_der



import os
import sys
import subprocess
import setuptools
from setuptools import setup, Extension


def detect_compiler():
    # Vérifier si un compilateur est spécifié dans les variables d'environnement
    compiler = os.environ.get('CC', None)
    if compiler:
        return compiler

    # gcc ?
    try:
        result = subprocess.run(['gcc', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        if result.returncode == 0:
            return 'gcc'
    except FileNotFoundError:
        pass

    # clang ?
    try:
        result = subprocess.run(['clang', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        if result.returncode == 0:
            return 'clang'
    except FileNotFoundError:
        pass

    # msvc ?
    if sys.platform == 'win32':
        try:
            result = subprocess.run(['cl'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
            if 'Microsoft (R) C/C++ Optimizing Compiler' in result.stderr:
                return 'msvc'
        except FileNotFoundError:
            pass

    return 'unknown compiler'

compiler = detect_compiler()
platform = sys.platform
print(f'compiling extension: {compiler} {platform}')

compile_args = ['-std=c11', '-O3', '-fopenmp']
link_args = []
if sys.platform == 'darwin':
    compile_args.extend(['-mmacosx-version-min=10.8', '-stdlib=libc++'])
    link_args.append('-lc++')


extensions = [
     Extension('nacl._libnacl',
               ['nacl/sparseutils.c',],
               extra_compile_args=compile_args,
               extra_link_args=link_args)]

setup(
    ext_modules=extensions,
)

# setup(name='sn-nacl',
#       #      packages=setuptools.find_packages(),
#       packages=['nacl', 'nacl.lib', 'nacl.models',
#                 'nacl.plotting', 'nacl.simulations',
#                 'nacl.data',
#                 'nacl.util'],
#       ext_modules=[Extension('nacl.lib._libnacl',
#                              ['nacl/lib/utils.c',],
#                              extra_compile_args=['-std=c11', '-O3'])],
#       version="0.3",
#       description="Supernova empirical model training package",
#       author='G. Augarde, N. Regnault, M. Betoule, S. Bongard',
#       author_email='nicolas.regnault@lpnhe.in2p3.fr',
#       #      package_dir={'': 'nacl'},
#       zip_safe=False,

#       install_requires=['numpy>=1.7', 'matplotlib', 'seaborn', 'cython', 'scikit-sparse', 'astropy', 'sparseqr'],
#       setup_requires=['pytest-runner'],
#       tests_require=['pytest-3'],
# )
